# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

"""
ScriptModule classes to compute the view of a tensor according to various different criteria.
"""

from typing import Optional, Tuple

import torch

import brevitas
from brevitas.core.function_wrapper import Identity
from brevitas.function.shape import over_batch_over_output_channels
from brevitas.function.shape import over_batch_over_tensor
from brevitas.function.shape import over_output_channels
from brevitas.function.shape import over_output_features
from brevitas.function.shape import over_tensor
from brevitas.utils.torch_utils import padding


class PermuteDims(brevitas.jit.ScriptModule):

    def __init__(self, permute_dims: Tuple[int, ...]) -> None:
        super(PermuteDims, self).__init__()
        self.permute_dims = permute_dims

    @brevitas.jit.script_method
    def forward(self, x: torch.Tensor):
        return x.permute(*self.permute_dims).contiguous()


class OverTensorView(brevitas.jit.ScriptModule):
    """
    ScriptMoodule to compute the :func:`~brevitas.function.shape.over_tensor` view of an input tensor.

    Examples:
        >>> view_module = OverTensorView()
        >>> y = view_module(torch.empty(size=[16, 6, 5, 5]))
        >>> y.shape
        torch.Size([2400])
    """

    def __init__(self) -> None:
        super(OverTensorView, self).__init__()

    @brevitas.jit.script_method
    def forward(self, x: torch.Tensor):
        shape = over_tensor(x)
        return x.reshape(shape)


class OverOutputChannelView(brevitas.jit.ScriptModule):
    """
    ScriptMoodule to compute the :func:`~brevitas.function.shape.over_output_channels` view of an
    input tensor.

    Examples:
        >>> view_module = OverOutputChannelView(permute_dims=None)
        >>> y = view_module(torch.empty(size=[16, 8, 5, 5]))
        >>> y.shape
        torch.Size([16, 200])
    """

    def __init__(self, permute_dims: Optional[Tuple[int, ...]] = None) -> None:
        super(OverOutputChannelView, self).__init__()
        if permute_dims is not None:
            self.permute_impl = PermuteDims(permute_dims)
        else:
            self.permute_impl = Identity()

    @brevitas.jit.script_method
    def forward(self, x: torch.Tensor):
        y = self.permute_impl(x)
        shape = over_output_channels(y)
        return y.reshape(shape)


class OverBatchOverTensorView(brevitas.jit.ScriptModule):
    """
    ScriptMoodule to compute the :func:`~brevitas.function.shape.over_batch_over_tensor` view of an
    input tensor.

    Examples:
        >>> view_module = OverBatchOverTensorView()
        >>> y = view_module(torch.empty(size=[8, 10, 5, 5]))
        >>> y.shape
        torch.Size([8, 250])
    """

    def __init__(self, permute_dims: Optional[Tuple[int, ...]] = None) -> None:
        super(OverBatchOverTensorView, self).__init__()
        if permute_dims is not None:
            self.permute_impl = PermuteDims(permute_dims)
        else:
            self.permute_impl = Identity()

    @brevitas.jit.script_method
    def forward(self, x: torch.Tensor):
        y = self.permute_impl(x)
        shape = over_batch_over_tensor(y)
        return y.reshape(shape)


class OverBatchOverOutputChannelView(brevitas.jit.ScriptModule):
    """
    ScriptModule to compute the :func:`~brevitas.function.shape.over_batch_over_output_channels`
    view of an input tensor.

    Examples:
        >>> view_module = OverBatchOverOutputChannelView()
        >>> y = view_module(torch.empty(size=[8, 10, 5, 5]))
        >>> y.shape
        torch.Size([8, 10, 25])
    """

    def __init__(self, permute_dims: Optional[Tuple[int, ...]] = None) -> None:
        super(OverBatchOverOutputChannelView, self).__init__()
        if permute_dims is not None:
            self.permute_impl = PermuteDims(permute_dims)
        else:
            self.permute_impl = Identity()

    @brevitas.jit.script_method
    def forward(self, x: torch.Tensor):
        y = self.permute_impl(x)
        shape = over_batch_over_output_channels(y)
        return y.reshape(shape)


class OverOutputFeaturesView(brevitas.jit.ScriptModule):
    """
    ScriptModule to compute the :func:`~brevitas.function.shape.over_output_features`
    view of an input tensor.

    Examples:
        >>> view_module = OverOutputFeaturesView()
        >>> y = view_module(torch.empty(size=[8, 10, 25]))
        >>> y.shape
        torch.Size([80, 25])
    """

    def __init__(self, permute_dims: Optional[Tuple[int, ...]] = None) -> None:
        super(OverOutputFeaturesView, self).__init__()
        if permute_dims is not None:
            self.permute_impl = PermuteDims(permute_dims)
        else:
            self.permute_impl = Identity()

    @brevitas.jit.script_method
    def forward(self, x: torch.Tensor):
        y = self.permute_impl(x)
        shape = over_output_features(y)
        return y.reshape(shape)


class OverSubChannelBlockView(brevitas.jit.ScriptModule):
    __constants__ = ['expanded_groupwise_shape', 'group_size', 'group_dim']

    def __init__(self, expanded_groupwise_shape, group_size, group_dim) -> None:
        super(OverSubChannelBlockView, self).__init__()
        self.expanded_groupwise_shape = expanded_groupwise_shape
        self.group_dim = group_dim
        self.group_size = group_size

    @brevitas.jit.script_method
    def forward(self, x: torch.Tensor):
        # This one is a bit tricky but we could end up here:
        # - If we quantize the zero point, which will already have expanded shape matching the scale (although no padding, but we don't need the padding)
        # - Groupwise HQO quantization, where weight will already have been padded and expanded
        if len(x.shape) == len(self.expanded_groupwise_shape):
            return x
        y = torch.nn.functional.pad(
            x, padding(x, self.group_size, self.group_dim), mode='constant', value=0.)
        y = y.view(self.expanded_groupwise_shape)
        return y


class DynamicOverSubChannelBlockView(brevitas.jit.ScriptModule):
    __constants__ = ['group_size', 'group_dim']

    def __init__(self, group_size, group_dim) -> None:
        super(DynamicOverSubChannelBlockView, self).__init__()
        self.group_size = group_size
        self.group_dim = group_dim

    @brevitas.jit.script_method
    def forward(self, x):

        tensor_shape = x.shape
        tensor_shape_list = list(tensor_shape)
        pad = padding(x, self.group_size, self.group_dim)

        x = torch.nn.functional.pad(x, pad, mode='constant', value=0.)

        tensor_shape = x.shape
        tensor_shape_list = list(tensor_shape)
        tensor_shape_list[self.group_dim] = int(tensor_shape_list[self.group_dim] / self.group_size)
        block_dim = self.group_dim + 1 if self.group_dim != -1 else -1
        tensor_shape_list.insert(block_dim, self.group_size)
        x = x.view(tensor_shape_list)
        return x


class StatsInputViewShapeImpl(object):
    """
    Enum-like object to collect pointers to variants of ScriptModules that perform a view on a tensor.
    All adhere to the same interface.
    """
    OVER_TENSOR = OverTensorView
    OVER_OUTPUT_CHANNELS = OverOutputChannelView
    OVER_BATCH_OVER_TENSOR = OverBatchOverTensorView
    OVER_BATCH_OVER_OUTPUT_CHANNELS = OverBatchOverOutputChannelView
    OVER_OUTPUT_FEATURES = OverOutputFeaturesView
    OVER_SUBCHANNEL_BLOCK = OverSubChannelBlockView
    DYNAMIC_OVER_SUBCHANNEL_BLOCK = DynamicOverSubChannelBlockView



"""
Copyright (c) 2016-     Facebook, Inc            (Adam Paszke)
Copyright (c) 2014-     Facebook, Inc            (Soumith Chintala)
Copyright (c) 2011-2014 Idiap Research Institute (Ronan Collobert)
Copyright (c) 2012-2014 Deepmind Technologies    (Koray Kavukcuoglu)
Copyright (c) 2011-2012 NEC Laboratories America (Koray Kavukcuoglu)
Copyright (c) 2011-2013 NYU                      (Clement Farabet)
Copyright (c) 2006-2010 NEC Laboratories America (Ronan Collobert, Leon Bottou, Iain Melvin, Jason Weston)
Copyright (c) 2006      Idiap Research Institute (Samy Bengio)
Copyright (c) 2001-2004 Idiap Research Institute (Ronan Collobert, Samy Bengio, Johnny Mariethoz)

All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

3. Neither the names of Xilinx, Facebook, Deepmind Technologies, NYU,
   NEC Laboratories America and IDIAP Research Institute nor the names
   of its contributors may be used to endorse or promote products derived
   from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.

Forked as-is from PyTorch 2.0.1
"""

import enum
import inspect
import numbers
import types
import typing
from typing import Any, Callable, cast, Dict, List, NamedTuple, Optional, Tuple, TYPE_CHECKING
import warnings

import torch
from torch._jit_internal import boolean_dispatched
from brevitas.backport._ops import OpOverload
from brevitas.backport._ops import OpOverloadPacket

from ._compatibility import compatibility

if TYPE_CHECKING:
    from .node import Argument

__all__ = [
    "ArgsKwargsPair",
    "check_for_mutable_operation",
    "get_signature_for_torch_op",
    "create_type_hint",
    "type_matches",
    "normalize_function",
    "normalize_module"]


@compatibility(is_backward_compatible=False)
class ArgsKwargsPair(NamedTuple):
    """
    Simple named tuple for wrapping args/kwargs pairs.
    """
    args: Tuple[Any, ...]
    kwargs: Dict[str, Any]


_manual_overrides: Dict[Callable, List[inspect.Signature]] = {}


def _nonzero_schemas():
    signatures = []

    def nonzero(self):
        pass

    signatures.append(inspect.signature(nonzero))

    def nonzero(self, *, as_tuple: bool):  # type: ignore[no-redef]
        pass

    signatures.append(inspect.signature(nonzero))

    return signatures


_manual_overrides[torch.nonzero] = _nonzero_schemas()


class _FakeGlobalNamespace:

    def __getattr__(self, name):
        if name == 'torch':
            return torch
        raise RuntimeError('Expected a torch namespace lookup')


_type_eval_globals = {
    'Tensor': torch.Tensor,
    'Device': torch.device,
    'Layout': torch.layout,
    'number': numbers.Number,
    'Future': torch.jit.Future,
    'AnyEnumType': enum.Enum,
    'QScheme': torch.qscheme,
    '__torch__': _FakeGlobalNamespace(),
    'NoneType': type(None),
    't': typing.TypeVar('t')}
for k in dir(typing):
    _type_eval_globals[k] = getattr(typing, k)


def _torchscript_type_to_python_type(ts_type: 'torch._C.JitType') -> Any:
    """
    Convert a TorchScript type to a Python type (including subtypes) via
    eval'ing the annotation_str. _type_eval_globals sets up expressions
    like "List" and "Future" to map to actual types (typing.List and jit.Future)
    """
    return eval(ts_type.annotation_str, _type_eval_globals)


def _torchscript_schema_to_signature(ts_schema: torch._C.FunctionSchema) -> inspect.Signature:
    from inspect import Parameter
    parameters: List[Parameter] = []
    for arg in ts_schema.arguments:
        arg_type = _torchscript_type_to_python_type(arg.type)
        default = arg.default_value if arg.has_default_value() else Parameter.empty
        # TODO: Figure out if this is safe. It seems like when generating the type signatures for
        # PythonArgParser, we emit signatures with `input` instead of `self` as the first tensor
        # argument name. Downstream, if someone converts that positional argument to a keyword
        # argument, the name mismatch will break things, so here we're going to normalize the
        # name to "input"
        name = arg.name if arg.name != 'self' else 'input'
        kind = Parameter.KEYWORD_ONLY if arg.kwarg_only else Parameter.POSITIONAL_OR_KEYWORD
        # "from" is a keyword therefore it must be a POSITIONAL_ONLY argument
        if name == "from":
            assert kind == Parameter.POSITIONAL_OR_KEYWORD
            # ParameterKind type is internal implementation detail to inspec package
            # which makes it hard to do type annoation
            kind = Parameter.POSITIONAL_ONLY  # type: ignore[assignment]
            # This renders all previous arguments to positional only
            for idx, p in enumerate(parameters):
                assert p.kind == Parameter.POSITIONAL_OR_KEYWORD
                parameters[idx] = Parameter(
                    name=p.name,
                    kind=Parameter.POSITIONAL_ONLY,
                    default=p.default,
                    annotation=p.annotation)
        parameters.append(Parameter(name=name, kind=kind, default=default, annotation=arg_type))
    return_types = [_torchscript_type_to_python_type(ret.type) for ret in ts_schema.returns]
    if len(return_types) == 0:
        return_type = None
    elif len(return_types) == 1:
        return_type = return_types[0]
    else:
        return_type = tuple(return_types)

    return inspect.Signature(parameters, return_annotation=return_type)


@compatibility(is_backward_compatible=False)
def check_for_mutable_operation(
        target: Callable, args: Tuple['Argument', ...], kwargs: Dict[str, 'Argument']):
    signatures, schemas = get_signature_for_torch_op(target, return_schemas=True)

    if signatures and schemas:
        matched_schemas = []

        # Iterate through all of the schema until we find one that matches
        # If one matches, populate `new_args_and_kwargs` with the new args/kwargs
        # values. If none matches, `new_args_and_kwargs` will be None
        for candidate_signature, schema in zip(signatures, schemas):
            try:
                candidate_signature.bind(*args, **kwargs)
                matched_schemas.append((candidate_signature, schema))
            except TypeError as e:
                continue

        def throw_if_mutable(schema):
            if schema.is_mutable:
                raise RuntimeError(
                    f'Tried to trace mutable operation {schema}. FX only supports functional '
                    f'code, so operations that mutate operands in-place (e.g. via `out` arguments) '
                    f'are not supported')

        if len(matched_schemas) == 0:
            # Did not match any schema. Cannot check for mutation
            pass
        elif len(matched_schemas) == 1:
            # Matched exactly one schema, unambiguous
            _, schema_to_check = matched_schemas[0]
            throw_if_mutable(schema_to_check)
            pass
        else:
            # Ambiguous schema match. Since mutability checking is best effort,
            # do nothing.
            pass


@compatibility(is_backward_compatible=False)
def get_signature_for_torch_op(op: Callable, return_schemas: bool = False):
    """
    Given an operator on the `torch` namespace, return a list of `inspect.Signature`
    objects corresponding to the overloads of that op.. May return `None` if a signature
    could not be retrieved.

    Args:
        op (Callable): An operator on the `torch` namespace to look up a signature for

    Returns:
        Optional[List[inspect.Signature]]: A list of signatures for the overloads of this
            operator, or None if the operator signatures could not be retrieved. If
            return_schemas=True, returns a tuple containing the optional Python signatures
            and the optional TorchScript Function signature
    """
    if isinstance(op, OpOverload):
        schemas = [op._schema]
    elif isinstance(op, OpOverloadPacket):
        schemas = [getattr(op, overload)._schema for overload in op.overloads()]
    else:
        override = _manual_overrides.get(op)
        if override:
            return (override, None) if return_schemas else None

        aten_fn = torch.jit._builtins._find_builtin(op)

        if aten_fn is None:
            return (None, None) if return_schemas else None
        schemas = torch._C._jit_get_schemas_for_operator(aten_fn)

    signatures = [_torchscript_schema_to_signature(schema) for schema in schemas]
    return (signatures, schemas) if return_schemas else signatures


@compatibility(is_backward_compatible=False)
def create_type_hint(x):
    try:
        if isinstance(x, (list, tuple)):
            # todo(chilli): Figure out the right way for mypy to handle this
            if isinstance(x, list):

                def ret_type(x):
                    return List[x]  # type: ignore[valid-type]
            else:

                def ret_type(x):
                    return Tuple[x, ...]

            if len(x) == 0:
                return ret_type(Any)
            base_type = x[0]
            for t in x:
                if issubclass(t, base_type):
                    continue
                elif issubclass(base_type, t):
                    base_type = t
                else:
                    return ret_type(Any)
            return ret_type(base_type)
    except Exception as e:
        # We tried to create a type hint for list but failed.
        warnings.warn(f"We were not able to successfully create type hint from the type {x}")
        pass
    return x


@compatibility(is_backward_compatible=False)
def type_matches(signature_type: Any, argument_type: Any):
    sig_origin_type = getattr(signature_type, '__origin__', signature_type)

    if signature_type is argument_type:
        return True

    # Union types in signature. Given type needs to match one of the
    # contained types in the Union
    if sig_origin_type is typing.Union and signature_type != argument_type:
        sig_contained = signature_type.__args__
        return any(type_matches(c, argument_type) for c in sig_contained)

    if signature_type is List[int] and argument_type is int:
        # int can be promoted to List[int]
        return True

    if getattr(signature_type, '__origin__', None) in {list, List}:
        sig_el_type = signature_type.__args__[0]
        if not inspect.isclass(sig_el_type):
            warnings.warn(
                f"Does not support nested parametric types, got {signature_type}. Please file a bug."
            )
            return False
        if getattr(argument_type, '__origin__', None) in {list, List}:
            return issubclass(argument_type.__args__[0], sig_el_type)

        def is_homogeneous_tuple(t):
            if not getattr(t, '__origin__', None) in {tuple, Tuple}:
                return False
            contained = t.__args__
            if t.__args__ == ((),):  # Tuple[()].__args__ == ((),) for some reason
                return True
            return all((c is Ellipsis) or issubclass(c, sig_el_type) for c in contained)

        # Tuple[T] is accepted for List[T] parameters
        return is_homogeneous_tuple(argument_type)

    # Dtype is an int in schemas
    if signature_type is int and argument_type is torch.dtype:
        return True

    if signature_type is numbers.Number and argument_type in {int, float}:
        return True
    if inspect.isclass(argument_type) and inspect.isclass(signature_type):
        return issubclass(argument_type, signature_type)

    return False


@compatibility(is_backward_compatible=False)
def normalize_function(
        target: Callable,
        args: Tuple[Any],
        kwargs: Optional[Dict[str, Any]] = None,
        arg_types: Optional[Tuple[Any]] = None,
        kwarg_types: Optional[Dict[str, Any]] = None,
        normalize_to_only_use_kwargs: bool = False) -> Optional[ArgsKwargsPair]:
    """
    Returns normalized arguments to PyTorch functions. This means that
    `args/kwargs` will be matched up to the functional's
    signature and return exclusively kwargs in positional order if
    `normalize_to_only_use_kwargs` is True.
    Also populates default values. Does not support positional-only
    parameters or varargs parameters (*args, **kwargs). Does not support modules.

    May require `arg_types` and `kwarg_types` in order to disambiguate overloads.

    Args:
        target (Callable): Function that we are normalizing
        args (Tuple[Any]): Tuple of args to the function
        kwargs (Optional[Dict[str, Any]]): Dict of kwargs to the function
        arg_types (Optional[Tuple[Any]]): Tuple of arg types for the args
        kwarg_types (Optional[Dict[str, Any]]): Dict of arg types for the kwargs
        normalize_to_only_use_kwargs (bool): Whether to normalize to only use kwargs.

    Returns:

        Returns normalized_args_and_kwargs, or `None` if not successful.
    """
    if kwargs is None:
        kwargs = {}
    new_args_and_kwargs = None
    if not isinstance(target, types.BuiltinFunctionType) and not (isinstance(
            target, (OpOverloadPacket, OpOverload))):
        target_for_analysis = target
        if target in boolean_dispatched:
            # HACK: `boolean_dispatch` as used in `torch.nn.functional` makes it so that we have
            # a 2-way dispatch based on a boolean value. Here we check that the `true` and `false`
            # branches of the dispatch have exactly the same signature. If they do, use the `true`
            # branch signature for analysis. Otherwise, leave this un-normalized
            assert not isinstance(target, str)
            dispatched = boolean_dispatched[target]
            if_true, if_false = dispatched['if_true'], dispatched['if_false']
            if inspect.signature(if_true).parameters != inspect.signature(if_false).parameters:
                return None
            target_for_analysis = if_true

        assert callable(target_for_analysis)
        sig = inspect.signature(inspect.unwrap(target_for_analysis))
        new_args_and_kwargs = _args_kwargs_to_normalized_args_kwargs(
            sig, args, kwargs, normalize_to_only_use_kwargs)
    else:
        assert callable(target)
        torch_op_schemas = get_signature_for_torch_op(target)
        matched_schemas = []
        if torch_op_schemas:
            # Iterate through all of the schema until we find one that matches
            # If one matches, populate `new_args_and_kwargs` with the new args/kwargs
            # values. If none matches, `new_args_and_kwargs` will be None
            for candidate_signature in torch_op_schemas:
                try:
                    candidate_signature.bind(*args, **kwargs)
                    matched_schemas.append(candidate_signature)
                except TypeError as e:
                    continue

            if len(matched_schemas) == 0:
                # Did not match any schema. Cannot normalize
                pass
            elif len(matched_schemas) == 1:
                # Matched exactly one schema, unambiguous
                new_args_and_kwargs = _args_kwargs_to_normalized_args_kwargs(
                    matched_schemas[0], args, kwargs, normalize_to_only_use_kwargs)
            else:
                if arg_types is not None or kwarg_types is not None:
                    arg_types = arg_types if arg_types else cast(Tuple[Any], ())
                    kwarg_types = kwarg_types if kwarg_types else {}
                    for candidate_signature in torch_op_schemas:
                        sig_matches = True
                        try:
                            bound_types = candidate_signature.bind(*arg_types, **kwarg_types)
                            for arg_name, arg_type in bound_types.arguments.items():
                                param = candidate_signature.parameters[arg_name]
                                sig_matches = sig_matches and type_matches(
                                    param.annotation, arg_type)
                        except TypeError as e:
                            sig_matches = False
                        if sig_matches:
                            new_args_and_kwargs = _args_kwargs_to_normalized_args_kwargs(
                                candidate_signature, args, kwargs, normalize_to_only_use_kwargs)
                            break
                else:
                    # Matched more than one schema. In this situation, the caller must provide the types of
                    # the arguments of the overload they expect.
                    schema_printouts = '\n'.join(str(schema) for schema in matched_schemas)
                    raise RuntimeError(
                        f'Tried to normalize arguments to {torch.typename(target)} but '
                        f'the schema match was ambiguous! Please provide argument types to '
                        f'the normalize_arguments() call. Available schemas:\n{schema_printouts}')

    return new_args_and_kwargs


@compatibility(is_backward_compatible=False)
def normalize_module(
        root: torch.nn.Module,
        target: str,
        args: Tuple[Any],
        kwargs: Optional[Dict[str, Any]] = None,
        normalize_to_only_use_kwargs: bool = False) -> Optional[ArgsKwargsPair]:
    """
    Returns normalized arguments to PyTorch modules. This means that
    `args/kwargs` will be matched up to the functional's
    signature and return exclusively kwargs in positional order if
    `normalize_to_only_use_kwargs` is True.
    Also populates default values. Does not support positional-only
    parameters or varargs parameters (*args, **kwargs).

    Args:
        root (nn.Module): root module upon which we query modules
        target (Callable): Function that we are normalizing
        args (Tuple[Any]): Tuple of args to the function
        kwargs (Optional[Dict[str, Any]]): Dict of kwargs to the function
        normalize_to_only_use_kwargs (bool): Whether to normalize to only use kwargs.

    Returns:

        Returns normalized_args_and_kwargs, or `None` if not successful.
    """
    try:
        submod = root.get_submodule(target)
    except AttributeError as e:
        raise RuntimeError(
            f"Tried to normalize node with target {target} but root did not "
            f"have that target!") from e
    if hasattr(submod.__class__, '__name__'):
        classname = submod.__class__.__name__
        if getattr(torch.nn, classname, None) == submod.__class__:
            sig = inspect.signature(inspect.unwrap(submod.forward))
            if kwargs is None:
                kwargs = {}
            new_args_and_kwargs = _args_kwargs_to_normalized_args_kwargs(
                sig, args, kwargs, normalize_to_only_use_kwargs)
            return new_args_and_kwargs
    return None


def _args_kwargs_to_normalized_args_kwargs(
        sig: inspect.Signature,
        args: Tuple[Any, ...],
        kwargs: Dict[str, Any],
        normalize_to_only_use_kwargs: bool) -> Optional[ArgsKwargsPair]:
    """
    Given a call target, args, and kwargs, return the arguments normalized into
    an ArgsKwargsPair, or None if the type signature is not supported by
    this normalization.

    Args:

        sig (inspect.Signature): Signature object for the target
        args (Tuple): Arguments that appear at the callsite for `target`
        kwargs (Dict): Keyword arguments that appear at the callsite for `target`
        normalize_to_only_use_kwargs (bool): Whether to normalize to only use kwargs.

    Returns:

        Optional[ArgsKwargsPair]: Normalized args and kwargs for `target`, or `None` if
            this target is not supported.
    """

    # Don't currently support positional-only
    # or varargs (*args, **kwargs) signatures
    supported_parameter_types = {
        inspect.Parameter.POSITIONAL_OR_KEYWORD, inspect.Parameter.KEYWORD_ONLY}
    if any(p.kind not in supported_parameter_types for p in sig.parameters.values()):
        # Add an exception for one signature, which is common for random/uniform, i.e.:
        # Tensor(a!) self, float from=0, float to=1, *, Generator? generator=None
        # `from` is Python keyword and as such functions with that signature should have
        # positional-only args, but at the same time they could be dispatched as kwargs
        if list(sig.parameters.keys()) != ['input', 'from', 'to', 'generator']:
            return None

    bound_args = sig.bind(*args, **kwargs)
    bound_args.apply_defaults()

    new_kwargs: Dict[str, Any] = {}
    new_args: List[Any] = []
    for i, param in enumerate(sig.parameters):
        if not normalize_to_only_use_kwargs and i < len(args):
            new_args.append(bound_args.arguments[param])
        else:
            new_kwargs[param] = bound_args.arguments[param]

    return ArgsKwargsPair(tuple(new_args), new_kwargs)



from brevitas.graph.calibrate import inference_mode
import brevitas.nn as qnn
from brevitas.quant.scaled_int import Int8ActPerTensorFloat
import torch
import torch.nn as nn

class Model(nn.Module):
    def __init__(self):
        super().__init__()
        self.linear = qnn.QuantLinear(3, 8, input_quant=Int8ActPerTensorFloat)
    
    def forward(self, x):
        return self.linear(x)
    

model = Model()
model.cuda()
model.eval()
input = torch.randn(1, 3, device='cuda')
with torch.no_grad(), inference_mode(model):
    model(input)
model = torch.compile(model)
with torch.no_grad():
    o = model(input)
def hook(*args, **kwargs):
    print("Oh")

model.register_forward_hook(hook)

print(o)


# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

import os

from setuptools import find_packages
from setuptools import setup

PROJECT_ROOT = os.path.dirname(os.path.abspath(__file__))
REQUIREMENTS_DIR = os.path.join(PROJECT_ROOT, 'requirements')


def read(*path):
    return open(os.path.join(*path), encoding='utf8').read()


def read_requirements(filename):
    return read(REQUIREMENTS_DIR, filename).splitlines()


setup(
    name="brevitas",
    use_scm_version=True,
    setup_requires=read_requirements('requirements-setup.txt'),
    description="Quantization-aware training in PyTorch",
    long_description=read(PROJECT_ROOT, 'README.md'),
    long_description_content_type="text/markdown",
    author="Alessandro Pappalardo",
    author_email="alessand@amd.com",
    url="https://github.com/Xilinx/brevitas",
    python_requires=">=3.8",
    install_requires=read_requirements('requirements.txt'),
    extras_require={
        "notebook": read_requirements('requirements-notebook.txt'),
        "dev": read_requirements('requirements-dev.txt'),
        "docs": read_requirements('requirements-docs.txt'),
        "export": read_requirements('requirements-export.txt'),
        "hadamard": read_requirements('requirements-hadamard.txt'),
        "test": read_requirements('requirements-test.txt'),
        "tts": read_requirements('requirements-tts.txt'),
        "stt": read_requirements('requirements-stt.txt'),
        "llm": read_requirements('requirements-llm.txt'),
        "vision": read_requirements('requirements-vision.txt'),
        "finn_integration": read_requirements('requirements-finn-integration.txt'),
        "ort_integration": read_requirements('requirements-ort-integration.txt')},
    packages=find_packages('src'),
    package_dir={'': 'src'},
    zip_safe=False,
    include_package_data=True,
    entry_points={
        'console_scripts': [
            'brevitas_bnn_pynq_train = brevitas_examples.bnn_pynq.bnn_pynq_train:main',
            'brevitas_qat_imagenet_val = brevitas_examples.imagenet_classification.qat.imagenet_val:main',
            'brevitas_quartznet_val = brevitas_examples.speech_to_text.quartznet_val:main',
            'brevitas_melgan_val = brevitas_examples.text_to_speech.melgan_val:main',
            'brevitas_quartznet_preprocess = brevitas_examples.speech_to_text.get_librispeech_data:main',
            'brevitas_melgan_preprocess = brevitas_examples.text_to_speech.preprocess_dataset:main',
            'brevitas_ptq_imagenet_val = brevitas_examples.imagenet_classification.ptq.ptq_evaluate:main',
            'brevitas_ptq_llm = brevitas_examples.llm.main:main'],})



# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

from dependencies import value

from brevitas.inject import ExtendedInjector
from brevitas.inject.enum import ScalingPerOutputType
from brevitas.quant.base import MSESymmetricScale
from brevitas.quant.experimental.float_base import FloatActBase
from brevitas.quant.experimental.float_base import FloatWeightBase
from brevitas.quant.experimental.float_base import Fp8e4m3Mixin
from brevitas.quant.experimental.float_base import Fp8e5m2Mixin
from brevitas.quant.experimental.float_base import ScaledFloatActBase
from brevitas.quant.experimental.float_base import ScaledFloatWeightBase


class FpFNUZMixin(ExtendedInjector):
    saturating = True

    @value
    def exponent_bias(exponent_bit_width):
        return 2 ** (exponent_bit_width - 1)


class FpFNUZWeight(FpFNUZMixin, FloatWeightBase):
    """
    FNUZ FP8 signed weight quantizer.
    """
    pass


class FpFNUZAct(FpFNUZMixin, FloatActBase):
    """
    FP8 signed activation quantizer.
    """
    pass


class FpFNUZWeightPerTensorFloat(FpFNUZMixin, ScaledFloatWeightBase):
    """
    FP8 signed weight quantizer with per-tensor absmax-based scaling.
    """
    scaling_per_output_type = ScalingPerOutputType.TENSOR


class FpFNUZActPerTensorFloat(FpFNUZMixin, ScaledFloatActBase):
    """
    FP8 signed activation quantizer with per-tensor static percentile-based scaling.
    """
    scaling_per_output_type = ScalingPerOutputType.TENSOR


class FpFNUZWeightPerChannelFloat(FpFNUZMixin, ScaledFloatWeightBase):
    """
    FP8 signed weight quantizer with per-channel absmax-based scaling.
    """
    scaling_per_output_type = ScalingPerOutputType.CHANNEL


class FpFNUZActPerChannelFloat2d(FpFNUZMixin, ScaledFloatActBase):
    """
    FP8 signed activation quantizer with per-channel static percentile-based scaling.
    """
    scaling_per_output_type = ScalingPerOutputType.CHANNEL
    scaling_stats_permute_dims = (1, 0, 2, 3)


class FpFNUZActPerTensorFloatMSE(FpFNUZMixin, MSESymmetricScale, ScaledFloatActBase):
    """
    FP8 signed activation quantizer with per-tensor static MSE-based scaling.
    """
    scaling_per_output_type = ScalingPerOutputType.TENSOR


class FpFNUZActPerChannelFloat2dMSE(FpFNUZMixin, MSESymmetricScale, ScaledFloatActBase):
    """
    FP8 signed activation quantizer with per-channel static MSE-based scaling.
    """
    scaling_per_output_type = ScalingPerOutputType.CHANNEL
    scaling_stats_permute_dims = (1, 0, 2, 3)


class FpFNUZWeightPerChannelFloatMSE(FpFNUZMixin, MSESymmetricScale, ScaledFloatWeightBase):
    """
    FP8 signed weight quantizer with per-channel MSE-based scaling.
    """
    scaling_per_output_type = ScalingPerOutputType.CHANNEL


class FpFNUZWeightPerTensorFloatMSE(FpFNUZMixin, MSESymmetricScale, ScaledFloatWeightBase):
    """
    FP8 signed weight quantizer with per-tensor MSE-based scaling.
    """
    scaling_per_output_type = ScalingPerOutputType.TENSOR


## Predefined FP8 Quantizers


class Fp8e4m3FNUZWeight(FpFNUZWeight, Fp8e4m3Mixin):
    """
    FNUZ FP8 E4M3 signed weight quantizer.
    """
    pass


class Fp8e4m3FNUZAct(FpFNUZAct, Fp8e4m3Mixin):
    """
    FNUZ FP8 E4M3 signed act quantizer.
    """
    pass


class Fp8e4m3FNUZWeightPerTensorFloat(FpFNUZWeightPerTensorFloat, Fp8e4m3Mixin):
    """
    FNUZ FP8 E4M3 per-tensor scaled signed weight quantizer.
    """
    pass


class Fp8e4m3FNUZWeightPerChannelFloat(FpFNUZWeightPerChannelFloat, Fp8e4m3Mixin):
    """
    FNUZ FP8 E4M3 per-channel scaled signed weight quantizer.
    """
    pass


class Fp8e4m3FNUZActPerTensorFloat(FpFNUZActPerTensorFloat, Fp8e4m3Mixin):
    """
    FNUZ FP8 E4M3 scaled signed act quantizer.
    """
    pass


class Fp8e4m3FNUZActPerTensorFloatMSE(FpFNUZActPerTensorFloatMSE, Fp8e4m3Mixin):
    """
    FNUZ FP8 E4M3 MSE-based scaled signed act quantizer.
    """
    pass


class Fp8e4m3FNUZWeightPerTensorFloatMSE(FpFNUZWeightPerTensorFloatMSE, Fp8e4m3Mixin):
    """
    FNUZ FP8 E4M3 MSE-based per-tensor scaled signed weight quantizer.
    """
    pass


class Fp8e4m3FNUZWeightPerChannelFloatMSE(FpFNUZWeightPerChannelFloatMSE, Fp8e4m3Mixin):
    """
    FNUZ FP8 E4M3 MSE-based per-channel scaled signed weight quantizer.
    """
    pass


class Fp8e5m2FNUZWeight(FpFNUZWeight, Fp8e5m2Mixin):
    """
    FNUZ FP8 e5m2 signed weight quantizer.
    """
    pass


class Fp8e5m2FNUZAct(FpFNUZAct, Fp8e5m2Mixin):
    """
    FNUZ FP8 e5m2 signed act quantizer.
    """
    pass


class Fp8e5m2FNUZWeightPerTensorFloat(FpFNUZWeightPerTensorFloat, Fp8e5m2Mixin):
    """
    FNUZ FP8 e5m2 per-tensor scaled signed weight quantizer.
    """
    pass


class Fp8e5m2FNUZWeightPerChannelFloat(FpFNUZWeightPerChannelFloat, Fp8e5m2Mixin):
    """
    FNUZ FP8 e5m2 per-channel scaled signed weight quantizer.
    """
    pass


class Fp8e5m2FNUZActPerTensorFloat(FpFNUZActPerTensorFloat, Fp8e5m2Mixin):
    """
    FNUZ FP8 e5m2 scaled signed act quantizer.
    """
    pass


class Fp8e5m2FNUZActPerTensorFloatMSE(FpFNUZActPerTensorFloatMSE, Fp8e5m2Mixin):
    """
    FNUZ FP8 e5m2 MSE-based scaled signed act quantizer.
    """
    pass


class Fp8e5m2FNUZWeightPerTensorFloatMSE(FpFNUZWeightPerTensorFloatMSE, Fp8e5m2Mixin):
    """
    FNUZ FP8 e5m2 MSE-based per-tensor scaled signed weight quantizer.
    """
    pass


class Fp8e5m2FNUZWeightPerChannelFloatMSE(FpFNUZWeightPerChannelFloatMSE, Fp8e5m2Mixin):
    """
    FNUZ FP8 e5m2 MSE-based per-channel scaled signed weight quantizer.
    """
    pass



import brevitas.nn as qnn
import torch
from tqdm import tqdm


class UInt4Tensor(torch.Tensor):
    @staticmethod
    def __new__(cls, elem, **kwargs):
        assert elem.dtype is torch.uint8
        assert not kwargs.get("requires_grad", False)
        kwargs["requires_grad"] = False
        return torch.Tensor._make_wrapper_subclass(
            cls, elem.shape, dtype=torch.uint8, **kwargs
        )

    def __init__(self, elem, **kwargs):
        self.elem = elem
    @classmethod
    def from_unpacked(cls, unpacked):
        uint5 = UInt4Tensor(unpacked)
        print(uint5)
        return uint5


    def __tensor_flatten__(self):
        return ["elem"], None

    @staticmethod
    def __tensor_unflatten__(flattened, meta, outer_size, outer_stride):
        assert meta is None
        elem = flattened["elem"]
        return UInt4Tensor(elem)
    __torch_function__ = torch._C._disabled_torch_function_impl


class PerChannelSymmetricWeightUInt4Tensor(UInt4Tensor):
    @staticmethod
    def __new__(cls, elem, scales, **kwargs):
        return super().__new__(cls, elem, **kwargs)

    def __init__(self, elem, scales, **kwargs):
        super().__init__(elem, **kwargs)

        self.scales = scales

    def __tensor_flatten__(self):
        return ["elem", "scales"], None
    @classmethod
    def __torch_dispatch__(cls, func, types, args, kwargs=None):
        return super().__torch_dispatch__(func, types, args, kwargs)
    @staticmethod
    def __tensor_unflatten__(flattened, meta, outer_size, outer_stride):
        assert meta is None
        elem = flattened["elem"]
        scales = flattened["scales"]
        return PerChannelSymmetricWeightUInt4Tensor(elem, scales)
    @classmethod
    def from_unpacked(cls, unpacked, scales):
        return cls(unpacked, scales)

    @classmethod
    def from_float(cls, w_fp32):
        quant = PerChannelSymmetricWeightUInt4Tensor.from_unpacked(
            unpacked=w_fp32, scales=torch.randn(1)
        )

        return quant
c = PerChannelSymmetricWeightUInt4Tensor.from_float(torch.tensor([0, 5, 1, 3], dtype=torch.uint8))

print(c)
raise

print("Start")
a = qnn.QuantIdentity()
params = list(a.parameters())
opt = torch.optim.SGD(a.parameters())
for _ in tqdm(range(500)):
    o = a(torch.randn(1, 3))
    loss = o.mean()
    loss.backward()
    if _ == 400:
        print(params[0].grad)
        raise
    opt.step()


# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

from abc import ABCMeta
from abc import abstractmethod
from typing import Optional, Type, Union

import torch
from torch import Tensor
from torch.nn import Module

from brevitas.quant_tensor import QuantTensor
from brevitas.utils.torch_utils import compute_channel_view_shape

from .mixin import *
from .utils import merge_bn
from .utils import rename_state_dict_by_prefix


class QuantNonLinearActLayer(QuantNonLinearActMixin, QuantInputMixin, QuantLayerMixin, Module):
    __metaclass__ = ABCMeta

    def __init__(
            self,
            act_impl: Optional[Type[Module]],
            passthrough_act: bool,
            input_quant: Optional[ActQuantType],
            act_quant: Optional[ActQuantType],
            return_quant_tensor: bool,
            **kwargs):
        Module.__init__(self)
        QuantLayerMixin.__init__(self, return_quant_tensor)
        QuantInputMixin.__init__(self, input_quant, **kwargs)
        QuantNonLinearActMixin.__init__(self, act_impl, passthrough_act, act_quant, **kwargs)

    @property
    def channelwise_separable(self) -> bool:
        return True

    @property
    def requires_export_handler(self):
        return self.input_quant.is_quant_enabled or self.act_quant.is_quant_enabled

    def forward(self, input: Union[Tensor, QuantTensor]):
        input = self.unpack_input(input)
        quant_input = self.input_quant(input)
        # shortcut execution through the export impl during export
        if self.export_mode:
            out = self.export_handler(quant_input)
            self._set_global_is_quant_layer(False)
            return out
        out = self.act_quant(quant_input)
        out = self.pack_output(out)
        return out

    def _load_from_state_dict(
            self, state_dict, prefix, local_metadata, strict, missing_keys, unexpected_keys,
            error_msgs):
        # for retrocompatibility
        rename_state_dict_by_prefix(prefix + 'act_quant_proxy', prefix + 'act_quant', state_dict)
        super(QuantNonLinearActLayer, self)._load_from_state_dict(
            state_dict, prefix, local_metadata, strict, missing_keys, unexpected_keys, error_msgs)


class QuantInputOutputLayer(QuantOutputMixin, QuantInputMixin, QuantLayerMixin):
    __metaclass__ = ABCMeta

    def __init__(
            self,
            input_quant: Optional[ActQuantType],
            output_quant: Optional[ActQuantType],
            tie_input_output_quant: bool,
            return_quant_tensor: bool,
            **kwargs):
        QuantLayerMixin.__init__(self, return_quant_tensor)
        QuantInputMixin.__init__(self, input_quant, **kwargs)
        QuantOutputMixin.__init__(self, output_quant, **kwargs)
        # we have to account for quantization being enabled through kwargs
        if tie_input_output_quant:
            if self.input_quant.is_quant_enabled and self.act_quant.is_quant_enabled:
                raise RuntimeError("Enable only input or output quant with tie_input_output=True")
            if self.input_quant.is_quant_enabled:
                self.output_quant = self.input_quant
            if self.act_quant.is_quant_enabled:
                self.input_quant = self.output_quant

    @property
    def requires_export_handler(self):
        return self.input_quant.is_quant_enabled or self.act_quant.is_quant_enabled


class QuantWeightBiasInputOutputLayer(QuantBiasMixin, QuantWeightMixin, QuantInputOutputLayer):
    __metaclass__ = ABCMeta

    def __init__(
            self,
            weight_quant: Optional[WeightQuantType],
            bias_quant: Optional[BiasQuantType],
            input_quant: Optional[ActQuantType],
            output_quant: Optional[ActQuantType],
            return_quant_tensor: bool,
            **kwargs):
        QuantInputOutputLayer.__init__(
            self,
            input_quant,
            output_quant,
            tie_input_output_quant=False,
            return_quant_tensor=return_quant_tensor,
            **kwargs)
        QuantWeightMixin.__init__(self, weight_quant, **kwargs)
        QuantBiasMixin.__init__(self, bias_quant, **kwargs)
        self._quant_load_model_mode = False

    @abstractmethod
    def inner_forward_impl(self, x: Tensor, quant_weight: Tensor, quant_bias: Optional[Tensor]):
        pass

    @abstractmethod
    def max_acc_bit_width(self, input_bit_width: Tensor, quant_weight_bit_width: Tensor):
        pass

    @property
    def requires_export_handler(self):
        return (
            self.input_quant.is_quant_enabled or self.weight_quant.is_quant_enabled or
            self.bias_quant.is_quant_enabled or self.output_quant.is_quant_enabled)

    @property
    def per_elem_ops(self):  # optional, so concrete impl + error if not overridden
        raise NotImplementedError

    def merge_bn_in(self, bn):
        merge_bn(self, bn, output_channel_dim=self.output_channel_dim)

    def forward_impl(self, inp: Union[Tensor, QuantTensor]) -> Union[Tensor, QuantTensor]:

        inp = self.unpack_input(inp)

        # shortcut execution through the export impl during export
        if self.export_mode:
            out = self.export_handler(inp)
            self._set_global_is_quant_layer(False)
            return out

        quant_input = self.input_quant(inp)
        quant_weight = self.quant_weight(quant_input)

        compute_output_quant_tensor = isinstance(quant_input, QuantTensor) and isinstance(
            quant_weight, QuantTensor)
        if not (compute_output_quant_tensor or
                self.output_quant.is_quant_enabled) and self.return_quant_tensor:
            raise RuntimeError("QuantLayer is not correctly configured")

        if self.bias is not None:
            quant_bias = self.bias_quant(self.bias, quant_input, quant_weight)
        else:
            quant_bias = None
        output_tensor = self.inner_forward_impl(quant_input, quant_weight, quant_bias)

        quant_output = self.output_quant(output_tensor)
        return self.pack_output(quant_output)

    def _load_from_state_dict(
            self, state_dict, prefix, local_metadata, strict, missing_keys, unexpected_keys,
            error_msgs):
        bias_key = prefix + 'bias'
        # If the state dict has a bias and the module does not, bias correction was used
        # We add a bias module to prevent failing during the load of the state dict
        if (bias_key in state_dict) and (self.bias is None) and self._quant_load_model_mode:
            self.register_parameter(
                'bias',
                torch.nn.Parameter(
                    torch.zeros(
                        self.out_channels, device=self.weight.device, dtype=self.weight.dtype)))
        super(QuantWeightBiasInputOutputLayer, self)._load_from_state_dict(
            state_dict, prefix, local_metadata, strict, missing_keys, unexpected_keys, error_msgs)



# Copyright (C) 2024, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

from functools import partial
from typing import Tuple, Type

import numpy as np
import torch
from torch import hub
from torch import Tensor
from torch.nn import Module
import torch.nn as nn
from torch.utils.data import DataLoader
from torch.utils.data import Dataset
from torch.utils.data import Subset
import torchvision
import torchvision.transforms as transforms
from tqdm import tqdm

from brevitas.core.scaling.pre_scaling import AccumulatorAwareParameterPreScaling
from brevitas.function import abs_binary_sign_grad
from brevitas.graph.calibrate import bias_correction_mode
from brevitas.graph.calibrate import calibration_mode

from .ep_init import apply_ep_init
from .quant import *
from .resnet import float_resnet18
from .resnet import quant_resnet18

__all__ = [
    "apply_ep_init",
    "apply_act_calibrate",
    "apply_bias_correction",
    "get_model_by_name",
    "filter_params",
    "create_calibration_dataloader",
    "get_cifar10_dataloaders",
    "train_for_epoch",
    "evaluate_topk_accuracies"]

device = 'cuda' if torch.cuda.is_available() else 'cpu'

model_impl = {
    "float_resnet18":
        float_resnet18,
    "quant_resnet18_w4a4_a2q_16b":
        partial(
            quant_resnet18,
            act_bit_width=4,
            acc_bit_width=16,
            weight_bit_width=4,
            weight_quant=CommonIntAccumulatorAwareWeightQuant),
    "quant_resnet18_w4a4_a2q_15b":
        partial(
            quant_resnet18,
            act_bit_width=4,
            acc_bit_width=15,
            weight_bit_width=4,
            weight_quant=CommonIntAccumulatorAwareWeightQuant),
    "quant_resnet18_w4a4_a2q_14b":
        partial(
            quant_resnet18,
            act_bit_width=4,
            acc_bit_width=14,
            weight_bit_width=4,
            weight_quant=CommonIntAccumulatorAwareWeightQuant),
    "quant_resnet18_w4a4_a2q_13b":
        partial(
            quant_resnet18,
            act_bit_width=4,
            acc_bit_width=13,
            weight_bit_width=4,
            weight_quant=CommonIntAccumulatorAwareWeightQuant),
    "quant_resnet18_w4a4_a2q_12b":
        partial(
            quant_resnet18,
            act_bit_width=4,
            acc_bit_width=12,
            weight_bit_width=4,
            weight_quant=CommonIntAccumulatorAwareWeightQuant),
    "quant_resnet18_w4a4_a2q_plus_16b":
        partial(
            quant_resnet18,
            act_bit_width=4,
            acc_bit_width=16,
            weight_bit_width=4,
            weight_quant=CommonIntAccumulatorAwareZeroCenterWeightQuant),
    "quant_resnet18_w4a4_a2q_plus_15b":
        partial(
            quant_resnet18,
            act_bit_width=4,
            acc_bit_width=15,
            weight_bit_width=4,
            weight_quant=CommonIntAccumulatorAwareZeroCenterWeightQuant),
    "quant_resnet18_w4a4_a2q_plus_14b":
        partial(
            quant_resnet18,
            act_bit_width=4,
            acc_bit_width=14,
            weight_bit_width=4,
            weight_quant=CommonIntAccumulatorAwareZeroCenterWeightQuant),
    "quant_resnet18_w4a4_a2q_plus_13b":
        partial(
            quant_resnet18,
            act_bit_width=4,
            acc_bit_width=13,
            weight_bit_width=4,
            weight_quant=CommonIntAccumulatorAwareZeroCenterWeightQuant),
    "quant_resnet18_w4a4_a2q_plus_12b":
        partial(
            quant_resnet18,
            act_bit_width=4,
            acc_bit_width=12,
            weight_bit_width=4,
            weight_quant=CommonIntAccumulatorAwareZeroCenterWeightQuant)}

root_url = 'https://github.com/Xilinx/brevitas/releases/download/'

model_url = {
    "float_resnet18":
        f"{root_url}/a2q_cifar10_r1/float_resnet18-1d98d23a.pth",
    "quant_resnet18_w4a4_a2q_12b":
        f"{root_url}/a2q_cifar10_r1/quant_resnet18_w4a4_a2q_12b-8a440436.pth",
    "quant_resnet18_w4a4_a2q_13b":
        f"{root_url}/a2q_cifar10_r1/quant_resnet18_w4a4_a2q_13b-8c31a2b1.pth",
    "quant_resnet18_w4a4_a2q_14b":
        f"{root_url}/a2q_cifar10_r1/quant_resnet18_w4a4_a2q_14b-267f237b.pth",
    "quant_resnet18_w4a4_a2q_15b":
        f"{root_url}/a2q_cifar10_r1/quant_resnet18_w4a4_a2q_15b-0d5bf266.pth",
    "quant_resnet18_w4a4_a2q_16b":
        f"{root_url}/a2q_cifar10_r1/quant_resnet18_w4a4_a2q_16b-d0af41f1.pth",
    "quant_resnet18_w4a4_a2q_plus_12b":
        f"{root_url}/a2q_cifar10_r1/quant_resnet18_w4a4_a2q_plus_12b-d69f003b.pth",
    "quant_resnet18_w4a4_a2q_plus_13b":
        f"{root_url}/a2q_cifar10_r1/quant_resnet18_w4a4_a2q_plus_13b-332aaf81.pth",
    "quant_resnet18_w4a4_a2q_plus_14b":
        f"{root_url}/a2q_cifar10_r1/quant_resnet18_w4a4_a2q_plus_14b-5a2d11aa.pth",
    "quant_resnet18_w4a4_a2q_plus_15b":
        f"{root_url}/a2q_cifar10_r1/quant_resnet18_w4a4_a2q_plus_15b-3c89551a.pth",
    "quant_resnet18_w4a4_a2q_plus_16b":
        f"{root_url}/a2q_cifar10_r1/quant_resnet18_w4a4_a2q_plus_16b-19973380.pth"}


def get_model_by_name(
        model_name: str,
        pretrained: bool = False,
        init_from_float_checkpoint: bool = False) -> nn.Module:

    assert model_name in model_impl, f"Error: {model_name} not implemented."
    assert not (pretrained and init_from_float_checkpoint), "Error: pretrained and init_from_float_checkpoint cannot both be true."
    model: Module = model_impl[model_name]()

    if init_from_float_checkpoint:
        checkpoint = model_url["float_resnet18"]
        state_dict = hub.load_state_dict_from_url(checkpoint, progress=True, map_location='cpu')
        model.load_state_dict(state_dict, strict=True)

    elif pretrained:
        checkpoint = model_url[model_name]
        state_dict = hub.load_state_dict_from_url(checkpoint, progress=True, map_location='cpu')
        if model_name.startswith("quant"):
            # fixes issue when bias keys are missing in the pre-trained state_dict when loading from checkpoint
            _prepare_bias_corrected_quant_model(model)
        model.load_state_dict(state_dict, strict=True)

    return model


def filter_params(named_params, decay):
    decay_params, no_decay_params = [], []
    for name, param in named_params:
        # Do not apply weight decay to the bias or any scaling parameters
        if 'scaling' in name or name.endswith(".bias"):
            no_decay_params.append(param)
        else:
            decay_params.append(param)
    return [{
        'params': no_decay_params, 'weight_decay': 0.}, {
            'params': decay_params, 'weight_decay': decay}]


def create_calibration_dataloader(
        dataset: Dataset, batch_size: int, num_workers: int, subset_size: int) -> DataLoader:

    all_indices = np.arange(len(dataset))
    cur_indices = np.random.choice(all_indices, size=subset_size)
    subset = Subset(dataset, cur_indices)
    loader = DataLoader(subset, batch_size=batch_size, num_workers=num_workers, pin_memory=True)
    return loader


def get_cifar10_dataloaders(
        data_root: str,
        batch_size_train: int = 128,
        batch_size_test: int = 100,
        num_workers: int = 2,
        pin_memory: bool = True,
        download: bool = False) -> Tuple[Type[DataLoader]]:

    mean, std = [0.491, 0.482, 0.447], [0.247, 0.243, 0.262]

    # create training dataloader
    transform_train = transforms.Compose([
        transforms.RandomCrop(32, padding=4),
        transforms.RandomHorizontalFlip(),
        transforms.ToTensor(),
        transforms.Normalize(mean=mean, std=std)])
    trainset = torchvision.datasets.CIFAR10(
        root=data_root,
        train=True,
        download=download,
        transform=transform_train,
    )
    trainloader = DataLoader(
        trainset,
        batch_size=batch_size_train,
        shuffle=True,
        num_workers=num_workers,
        pin_memory=pin_memory,
        persistent_workers=num_workers > 0,
    )

    # creating the validation dataloader
    transform_test = transforms.Compose([
        transforms.ToTensor(),
        transforms.Normalize(mean=mean, std=std),])
    testset = torchvision.datasets.CIFAR10(
        root=data_root,
        train=False,
        download=download,
        transform=transform_test,
    )
    testloader = DataLoader(
        testset,
        batch_size=batch_size_test,
        shuffle=False,
        num_workers=num_workers,
        pin_memory=pin_memory,
        persistent_workers=num_workers > 0,
    )

    return trainloader, testloader


def apply_act_calibrate(calib_loader, model):
    model.eval()
    dtype = next(model.parameters()).dtype
    device = next(model.parameters()).device
    with torch.no_grad():
        with calibration_mode(model):
            for images, _ in tqdm(calib_loader):
                images = images.to(device)
                images = images.to(dtype)
                model(images)


def apply_bias_correction(calib_loader, model: nn.Module):
    model.eval()
    dtype = next(model.parameters()).dtype
    device = next(model.parameters()).device
    with torch.no_grad():
        with bias_correction_mode(model):
            for (images, _) in tqdm(calib_loader):
                images = images.to(device)
                images = images.to(dtype)
                model(images)


def _prepare_bias_corrected_quant_model(model: nn.Module):
    model.eval()
    dtype = next(model.parameters()).dtype
    device = next(model.parameters()).device
    images = torch.randn(10, 3, 32, 32)
    images = images.to(device)
    images = images.to(dtype)
    with torch.no_grad():
        with bias_correction_mode(model):
            model(images)


def train_for_epoch(trainloader, model, criterion, optimizer, reg_weight: float = 1e-3):
    model.train()
    model = model.to(device)

    tot_loss, reg_penalty = 0., 0.

    def acc_reg_penalty(module: AccumulatorAwareParameterPreScaling, inp, output):
        """Accumulate the regularization penalty across constrained layers"""
        nonlocal reg_penalty
        (weights, input_bit_width, input_is_signed) = inp
        s = module.scaling_impl(weights)  # s
        g = abs_binary_sign_grad(module.restrict_clamp_scaling(module.value))  # g
        T = module.calc_max_l1_norm(input_bit_width, input_is_signed)  # T / s
        cur_penalty = torch.relu(g - (T * s)).sum()
        reg_penalty += cur_penalty
        return output

    # Register a forward hook to accumulate the regularization penalty
    hook_fns = list()
    for mod in model.modules():
        if isinstance(mod, AccumulatorAwareParameterPreScaling):
            hook = mod.register_forward_hook(acc_reg_penalty)
            hook_fns.append(hook)

    progress_bar = tqdm(trainloader)
    for _, (images, targets) in enumerate(progress_bar):
        optimizer.zero_grad()
        images = images.to(device)
        targets = targets.to(device)
        outputs = model(images)
        task_loss: Tensor = criterion(outputs, targets)
        loss = task_loss + (reg_weight * reg_penalty)
        loss.backward()
        optimizer.step()
        reg_penalty = 0.  # reset the accumulated regularization penalty
        tot_loss += task_loss.item() * images.size(0)

    # Remove the registered forward hooks before exiting
    for hook in hook_fns:
        hook.remove()

    avg_loss = tot_loss / len(trainloader.dataset)
    return avg_loss


@torch.no_grad()
def evaluate_topk_accuracies(testloader, model, criterion):
    model.eval()
    model = model.to(device)

    progress_bar = tqdm(testloader)

    top_1, top_5, tot_loss = 0., 0., 0.
    for _, (images, targets) in enumerate(progress_bar):
        images = images.to(device)
        targets = targets.to(device)
        outputs: Tensor = model(images)
        loss: Tensor = criterion(outputs, targets)

        # Evaluating Top-1 and Top-5 accuracy
        _, y_pred = outputs.topk(5, 1, True, True)
        y_pred = y_pred.t()
        correct = y_pred.eq(targets.view(1, -1).expand_as(y_pred))
        top_1 += correct[0].float().sum().item()
        top_5 += correct.float().sum().item()
        tot_loss += loss.item() * images.size(0)
    top_1 /= len(testloader.dataset)
    top_5 /= len(testloader.dataset)
    tot_loss /= len(testloader.dataset)
    return top_1, top_5, tot_loss



# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

import argparse
from functools import partial
from itertools import product
import os
import random
from types import SimpleNamespace

import numpy as np
import pandas as pd
import torch
import torch.backends.cudnn as cudnn
import torch.nn.parallel
import torch.optim
import torch.utils.data
import torch.utils.data.distributed

from brevitas import __version__ as brevitas_version
from brevitas import config
from brevitas import torch_version
from brevitas.graph.quantize import preprocess_for_quantize
from brevitas.graph.target.flexml import preprocess_for_flexml_quantize
from brevitas_examples.imagenet_classification.ptq.ptq_common import apply_act_equalization
from brevitas_examples.imagenet_classification.ptq.ptq_common import apply_bias_correction
from brevitas_examples.imagenet_classification.ptq.ptq_common import apply_gpfq
from brevitas_examples.imagenet_classification.ptq.ptq_common import apply_gptq
from brevitas_examples.imagenet_classification.ptq.ptq_common import apply_learned_round_learning
from brevitas_examples.imagenet_classification.ptq.ptq_common import calibrate
from brevitas_examples.imagenet_classification.ptq.ptq_common import quantize_model
from brevitas_examples.imagenet_classification.ptq.utils import get_gpu_index
from brevitas_examples.imagenet_classification.ptq.utils import get_model_config
from brevitas_examples.imagenet_classification.ptq.utils import get_torchvision_model
from brevitas_examples.imagenet_classification.utils import generate_dataloader
from brevitas_examples.imagenet_classification.utils import SEED
from brevitas_examples.imagenet_classification.utils import validate

config.IGNORE_MISSING_KEYS = True


def parse_type(v, default_type):
    if v == 'None':
        return None
    else:
        return default_type(v)


def parse_bool(v):
    if isinstance(v, bool):
        return v
    if v.lower() in ('yes', 'true', 't', 'y'):
        return True
    elif v.lower() in ('no', 'false', 'f', 'n'):
        return False
    else:
        raise argparse.ArgumentTypeError('Boolean value expected.')


class hashabledict(dict):

    def __hash__(self):
        return hash(tuple(sorted(self.items())))


def unique(sequence):
    seen = set()
    return [x for x in sequence if not (x in seen or seen.add(x))]


# Torchvision models with top1 accuracy
TORCHVISION_TOP1_MAP = {
    'resnet18': 69.758,
    'mobilenet_v2': 71.898,
    'vit_b_32': 75.912,}

OPTIONS_DEFAULT = {
    'model_name': list(TORCHVISION_TOP1_MAP.keys()),
    'quant_format': ['int'],  # Quantization type (INT vs Float)
    'target_backend': ['layerwise'],  # Target backend
    'scale_factor_type': ['float_scale'],  # Scale factor type
    'weight_mantissa_bit_width': [4],
    'weight_exponent_bit_width': [3],
    'weight_narrow_range': [False],
    'layerwise_first_last_bit_width': [8],  # Input and weights bit width for first and last layer
    'act_mantissa_bit_width': [4],
    'act_exponent_bit_width': [3],
    'weight_bit_width': [8],  # Weight Bit Width
    'act_bit_width': [8],  # Act bit width
    'bias_bit_width': [32],  # Bias Bit-Width for Po2 scale
    'weight_quant_granularity': ['per_channel'],  # Scaling Per Output Channel
    'act_quant_granularity': ['per_tensor'],  # Scaling Per Output Channel
    'act_quant_type': ['sym'],  # Act Quant Type
    'act_scale_computation_type': ['static'],  # Act Quant Type
    'act_param_method': ['stats'],  # Act Param Method
    'weight_param_method': ['mse'],  # Weight Quant Type
    'bias_corr': [True],  # Bias Correction
    'graph_eq_iterations': [20],  # Graph Equalization
    'graph_eq_merge_bias': [True],  # Merge bias for Graph Equalization
    'act_equalization': ['layerwise'],  # Perform Activation Equalization (Smoothquant)
    'learned_round': [False],  # Enable/Disable Learned Round
    'gptq': [False],  # Enable/Disable GPTQ
    'gpfq': [False],  # Enable/Disable GPFQ
    'gpfa2q': [False],  # Enable/Disable GPFA2Q
    'gpfq_p': [1.0],  # GPFQ P
    'gpxq_act_order': [False],  # Use act_order euristics for GPxQ
    'accumulator_bit_width': [16],  # Accumulator bit width, only in combination with GPFA2Q
    'act_quant_percentile': [99.999],  # Activation Quantization Percentile
    'uint_sym_act_for_unsigned_values': [True],  # Whether to use unsigned act quant when possible
    'channel_splitting_ratio': [0.0],  # Channel Splitting ratio, 0.0 means no splitting
    'split_input': [True],  # Whether to split the input channels when applying channel splitting
    'merge_bn': [True]}  # Whether to merge BN layers

parser = argparse.ArgumentParser(description='PyTorch ImageNet PTQ Validation')
parser.add_argument('idx', type=int)
parser.add_argument(
    '--calibration-dir',
    default='/scratch/datasets/imagenet_symlink/calibration',
    help='path to folder containing Imagenet calibration folder')
parser.add_argument(
    '--validation-dir',
    default='/scratch/datasets/imagenet_symlink/val',
    help='path to folder containing Imagenet validation folder')

parser.add_argument('--workers', default=8, type=int, help='Number of data loading workers')
parser.add_argument(
    '--batch-size-calibration', default=64, type=int, help='Minibatch size for calibration')
parser.add_argument(
    '--batch-size-validation', default=256, type=int, help='Minibatch size for validation')
parser.add_argument('--calibration-samples', default=1000, type=int, help='Calibration size')
for option_name, option_value in OPTIONS_DEFAULT.items():
    if isinstance(option_value[0], bool):
        type_args = parse_bool
    else:
        type_args = partial(parse_type, default_type=type(option_value[0]))
    parser.add_argument(f'--{option_name}', default=option_value, nargs="+", type=type_args)


def main():
    args = parser.parse_args()
    random.seed(SEED)
    np.random.seed(SEED)
    torch.manual_seed(SEED)

    args.gpu = get_gpu_index(args.idx)
    print("Iter {}, GPU {}".format(args.idx, args.gpu))

    try:
        ptq_torchvision_models(args)
    except Exception as E:
        print("Exception at index {}: {}".format(args.idx, E))


def ptq_torchvision_models(args):
    # Generate all possible combinations, including invalid ones

    options = {k: getattr(args, k) for k, _ in OPTIONS_DEFAULT.items()}

    combinations = list(product(*options.values()))

    configs = []
    for combination in combinations:
        config_namespace = SimpleNamespace(
            **{k: v for k, v in zip(OPTIONS_DEFAULT.keys(), combination)})
        config_namespace = validate_config(config_namespace)
        if config_namespace.is_valid:
            configs.append(hashabledict(**config_namespace.__dict__))

    configs = unique(configs)

    if args.idx > len(configs) - 1:
        return

    config_namespace = SimpleNamespace(**configs[args.idx])
    print(config_namespace)

    fp_accuracy = TORCHVISION_TOP1_MAP[config_namespace.model_name]
    # Get model-specific configurations about input shapes and normalization
    model_config = get_model_config(config_namespace.model_name)

    # Generate calibration and validation dataloaders
    resize_shape = model_config['resize_shape']
    center_crop_shape = model_config['center_crop_shape']
    inception_preprocessing = model_config['inception_preprocessing']
    calib_loader = generate_dataloader(
        args.calibration_dir,
        args.batch_size_calibration,
        args.workers,
        resize_shape,
        center_crop_shape,
        args.calibration_samples,
        inception_preprocessing)
    val_loader = generate_dataloader(
        args.validation_dir,
        args.batch_size_validation,
        args.workers,
        resize_shape,
        center_crop_shape,
        inception_preprocessing=inception_preprocessing)

    # Get the model from torchvision
    model = get_torchvision_model(config_namespace.model_name)

    # Preprocess the model for quantization
    if config_namespace.target_backend == 'flexml':
        # Flexml requires static shapes, thus representative input is passed in
        img_shape = model_config['center_crop_shape']
        model = preprocess_for_flexml_quantize(
            model,
            torch.ones(1, 3, img_shape, img_shape),
            equalize_iters=config_namespace.graph_eq_iterations,
            equalize_merge_bias=config_namespace.graph_eq_merge_bias)
    elif config_namespace.target_backend == 'fx' or config_namespace.target_backend == 'layerwise':
        model = preprocess_for_quantize(
            model,
            equalize_iters=config_namespace.graph_eq_iterations,
            equalize_merge_bias=config_namespace.graph_eq_merge_bias,
            merge_bn=config_namespace.merge_bn,
            channel_splitting_ratio=config_namespace.channel_splitting_ratio,
            channel_splitting_split_input=config_namespace.split_input)
    else:
        raise RuntimeError(f"{config_namespace.target_backend} backend not supported.")

    if config_namespace.act_equalization is not None:
        print("Applying activation equalization:")
        apply_act_equalization(
            model, calib_loader, layerwise=config_namespace.act_equalization == 'layerwise')

    # Define the quantized model
    quant_model = quantize_model(
        model,
        quant_format=config_namespace.quant_format,
        backend=config_namespace.target_backend,
        act_bit_width=config_namespace.act_bit_width,
        layerwise_first_last_bit_width=config_namespace.layerwise_first_last_bit_width,
        weight_narrow_range=config_namespace.weight_narrow_range,
        weight_mantissa_bit_width=config_namespace.weight_mantissa_bit_width,
        weight_exponent_bit_width=config_namespace.weight_exponent_bit_width,
        act_mantissa_bit_width=config_namespace.act_mantissa_bit_width,
        act_exponent_bit_width=config_namespace.act_exponent_bit_width,
        weight_bit_width=config_namespace.weight_bit_width,
        weight_param_method=config_namespace.weight_param_method,
        act_param_method=config_namespace.act_param_method,
        bias_bit_width=config_namespace.bias_bit_width,
        act_scale_computation_type=config_namespace.act_scale_computation_type,
        weight_quant_granularity=config_namespace.weight_quant_granularity,
        act_quant_granularity=config_namespace.act_quant_granularity,
        act_quant_percentile=config_namespace.act_quant_percentile,
        act_quant_type=config_namespace.act_quant_type,
        scale_factor_type=config_namespace.scale_factor_type,
        uint_sym_act_for_unsigned_values=config_namespace.uint_sym_act_for_unsigned_values)

    # If available, use the selected GPU
    if args.gpu is not None:
        torch.cuda.set_device(args.gpu)
        quant_model = quant_model.cuda(args.gpu)
        cudnn.benchmark = False

    # Calibrate the quant_model on the calibration dataloader
    print("Starting calibration")
    calibrate(calib_loader, quant_model)

    if config_namespace.gpfq:
        print("Performing GPFQ:")
        apply_gpfq(
            calib_loader,
            quant_model,
            p=config_namespace.gpfq_p,
            act_order=config_namespace.gpxq_act_order)

    if config_namespace.gpfa2q:
        print("Performing GPFA2Q:")
        apply_gpfq(
            calib_loader,
            quant_model,
            p=config_namespace.gpfq_p,
            act_order=config_namespace.gpxq_act_order,
            gpfa2q=config_namespace.gpfa2q,
            accumulator_bit_width=config_namespace.accumulator_bit_width)

    if config_namespace.gptq:
        print("Performing gptq")
        apply_gptq(calib_loader, quant_model, config_namespace.gpxq_act_order)

    if config_namespace.learned_round:
        print("Applying Learned Round:")
        apply_learned_round_learning(quant_model, calib_loader)

    if config_namespace.bias_corr:
        print("Applying bias correction")
        apply_bias_correction(calib_loader, quant_model)

    # Validate the quant_model on the validation dataloader
    print("Starting validation")
    top1 = validate(val_loader, quant_model)

    # Generate metrics for benchmarking
    top1 = np.around(top1, decimals=3)
    acc_diff = np.around(top1 - fp_accuracy, decimals=3)
    acc_ratio = np.around(top1 / fp_accuracy, decimals=3)

    column_names = [k.replace('_', ' ').capitalize() for k in config_namespace.__dict__.keys()] + [
        'Top 1% floating point accuracy',
        'Top 1% quant accuracy',
        'Floating point accuracy - quant accuracy',
        'Quant accuracy / floating point accuracy',
        'Calibration size',
        'Calibration batch size',
        'Torch version',
        'Brevitas version']
    values = [v for _, v in config_namespace.__dict__.items()] + [
        fp_accuracy,
        top1,
        acc_diff,
        acc_ratio,
        args.calibration_samples,
        args.batch_size_calibration,
        torch_version,
        brevitas_version]
    torchvision_df = pd.DataFrame([values], columns=column_names)

    folder = './multirun/' + str(args.idx)
    os.makedirs(folder, exist_ok=True)
    torchvision_df.to_csv(os.path.join(folder, 'RESULTS_TORCHVISION.csv'), index=False)


def validate_config(config_namespace):
    is_valid = True
    # Flexml supports only per-tensor scale factors, power of two scale factors
    if config_namespace.target_backend == 'flexml' and (
            config_namespace.weight_quant_granularity == 'per_channel' or
            config_namespace.scale_factor_type == 'float_scale'):
        is_valid = False
    # Merge bias can be enabled only when graph equalization is enabled
    if config_namespace.graph_eq_iterations == 0 and config_namespace.graph_eq_merge_bias:
        is_valid = False
    # For fx and layerwise backend, we only test for bias with bit width equals to 32
    if (config_namespace.target_backend == 'fx' or config_namespace.target_backend
            == 'layerwise') and config_namespace.bias_bit_width == 16:
        is_valid = False
    # Only one of GPTQ, GPFQ, or GPA2Q can be enabled, or none
    multiple_gpxqs = float(config_namespace.gpfq) + float(config_namespace.gptq) + float(
        config_namespace.gpfa2q)
    if multiple_gpxqs > 1:
        is_valid = False
    elif multiple_gpxqs == 0:
        # no gpxq algorithm, set act order to None
        config_namespace.gpxq_act_order = None

    if config_namespace.act_equalization == 'layerwise' and config_namespace.target_backend == 'fx':
        is_valid = False
    if config_namespace.act_bit_width < config_namespace.weight_bit_width:
        is_valid = False

    if config_namespace.act_param_method == 'mse':
        config_namespace.act_quant_percentile = None
    # gpfq_p is needed for GPFQ and GPFA2Q
    if not config_namespace.gpfq and not config_namespace.gpfa2q:
        config_namespace.gpfq_p = None
    # accumulator bit width is not needed when not GPFA2Q
    if not config_namespace.gpfa2q:
        config_namespace.accumulator_bit_width = None

    if config_namespace.quant_format == 'int':
        config_namespace.weight_mantissa_bit_width = None
        config_namespace.weight_exponent_bit_width = None
        config_namespace.act_mantissa_bit_width = None
        config_namespace.act_exponent_bit_width = None

    if config_namespace.quant_format == 'float':
        config_namespace.act_quant_type = 'sym'
        config_namespace.weight_quant_type = 'sym'

    if config_namespace.quant_format == 'float':
        if config_namespace.weight_exponent_bit_width + config_namespace.weight_mantissa_bit_width != config_namespace.weight_bit_width - 1:
            is_valid = False
        if config_namespace.act_exponent_bit_width + config_namespace.act_mantissa_bit_width != config_namespace.act_bit_width - 1:
            is_valid = False
    # if channel splitting is disabled, no need for split input
    if not config_namespace.channel_splitting_ratio:
        config_namespace.split_input = None

    config_namespace.is_valid = is_valid
    return config_namespace


if __name__ == '__main__':
    main()



# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

import torch

from brevitas.function.ops_ste import round_ste
from brevitas.quant_tensor import _unpack_quant_tensor
from brevitas.quant_tensor.base_quant_tensor import GroupwisIntQuantTensorBase
from brevitas.quant_tensor.base_quant_tensor import QuantTensor
from brevitas.utils.torch_utils import float_internal_scale

from .int_torch_handler import INT_QUANT_TENSOR_FN_HANDLER
from .torch_handler import QUANT_TENSOR_FN_HANDLER

IS_VALID_ATOL = 2e-1
BFLOAT16_IS_VALID_ATOL = 0.5


class GroupwiseIntQuantTensor(GroupwisIntQuantTensorBase, QuantTensor):

    def __new__(cls, value, scale, zero_point, group_size, group_dim, bit_width, signed, training):

        if not isinstance(scale, torch.Tensor):
            scale = torch.tensor(scale, dtype=torch.float)
        if not isinstance(zero_point, torch.Tensor):
            zero_point = torch.tensor(zero_point, dtype=torch.float)
        if not isinstance(bit_width, torch.Tensor):
            bit_width = torch.tensor(bit_width, dtype=torch.float)
        if not isinstance(signed, torch.Tensor):
            signed = torch.tensor(signed, dtype=torch.bool)
        if not isinstance(training, torch.Tensor):
            training = torch.tensor(training, dtype=torch.bool)
        quant_tensor = super().__new__(
            cls, value, scale, zero_point, group_size, group_dim, bit_width, signed, training)
        return quant_tensor

    @property
    def signed(self):
        return self.signed_t.item()

    @property
    def training(self):
        return self.training_t.item()

    @property
    def saturating(self):
        return self.saturating_t.item()

    def __torch_function__(self, func, types, args=(), kwargs=None):
        if kwargs is None:
            kwargs = {}

        if func in QUANT_TENSOR_FN_HANDLER:
            return QUANT_TENSOR_FN_HANDLER[func](*args, **kwargs)
        else:
            args = _unpack_quant_tensor(args)
            kwargs = _unpack_quant_tensor(kwargs)
            return func(*args, **kwargs)

    def expand(self):
        curr_shape = self.value_.shape
        start_dim = self.group_dim if self.group_dim != -1 else -2
        new_value = self.value_.flatten(start_dim, start_dim + 1)
        if self.scale_.shape != ():
            new_scale = self.scale_.expand(curr_shape).flatten(start_dim, start_dim + 1)
        else:
            new_scale = self.scale_
        if self.zero_point_.shape != ():
            new_zp = self.zero_point_.expand(curr_shape).flatten(start_dim, start_dim + 1)
        else:
            new_zp = self.zero_point_

        return new_value, new_scale, new_zp

    @staticmethod
    def from_expanded(value, group_size, group_dim, compress=False):
        group_dim = group_dim if group_dim != -1 else -2
        size = list(value.shape)
        assert size[group_dim] % group_size == 0, 'Input channel is not divisible by group size'
        if compress:
            size[group_dim] = 1
        else:
            size[group_dim] = size[group_dim] // group_size
        size.insert(group_dim + 1, group_size)
        new_value = value.view(size)
        return new_value

    @property
    def tensor(self):
        return self.value

    @property
    def value(self):
        new_value, new_scale, new_zp = self.expand()
        return new_value

    @property
    def scale(self):
        new_value, new_scale, new_zp = self.expand()
        return new_scale

    @property
    def zero_point(self):
        new_value, new_scale, new_zp = self.expand()
        return new_zp

    @property
    def _pre_round_int_value(self):
        value = self.value
        scale = self.scale
        zero_point = self.zero_point
        if self.scale.dtype == torch.bfloat16:
            value = self.value.type(torch.float32)
            scale = self.scale.type(torch.float32)
            zero_point = self.zero_point.type(torch.float32)
        int_value = value / scale
        int_value = int_value + zero_point
        return int_value

    @property
    def is_valid(self):
        with torch.no_grad():
            pre_round_int_value = self._pre_round_int_value
            rounded_int_value = torch.round(pre_round_int_value)
            max_abs_diff = torch.max(torch.abs(pre_round_int_value - rounded_int_value))
            atol = BFLOAT16_IS_VALID_ATOL if self.value.dtype == torch.bfloat16 else IS_VALID_ATOL
            is_int = max_abs_diff < atol
            if self.bit_width >= 2:
                if self.signed:
                    is_upper_b = (2.0 ** (self.bit_width - 1) - 1 >= rounded_int_value).all()
                    is_lower_b = (-2.0 ** (self.bit_width - 1) <= rounded_int_value).all()
                else:
                    is_upper_b = (2.0 ** self.bit_width - 1 >= rounded_int_value).all()
                    is_lower_b = (0. <= rounded_int_value).all()
                return (is_int & is_upper_b & is_lower_b).item()
            else:  # binary case
                unique_vals = rounded_int_value.unique(
                    sorted=False, return_counts=False, return_inverse=False)
                is_binary = unique_vals.view(-1).size()[0] == 2
                is_signed = (unique_vals < 0.).any().item()
                sign_match = is_signed == self.signed
                return is_int.item() and is_binary and sign_match

    @property
    def device(self):
        value_device = self.value_.device
        is_same_device = True
        for t in [self.scale,
                  self.zero_point,
                  self.exponent_bit_width,
                  self.mantissa_bit_width,
                  self.exponent_bias]:
            is_same_device &= value_device == t.device
        if not is_same_device:
            raise RuntimeError("Value and metadata are on different devices")
        return value_device

    def int(self, float_datatype=False):
        if self.is_valid:
            int_value = round_ste(self._pre_round_int_value)
            if float_datatype:
                # Values at 8bit and lower can be represented exactly with float16 and bfloat16
                # otherwise (e.g. Int16 bias), we upscale to float32
                if self.bit_width <= 8.:
                    return int_value.type(self.scale.dtype)
                else:
                    return int_value.type(torch.float32)
            else:
                if self.bit_width <= 8. and self.signed_t.item():
                    return int_value.to(torch.int8)
                elif self.bit_width <= 8. and not self.signed_t.item():
                    return int_value.to(torch.uint8)
                else:
                    return int_value.to(torch.int32)
        else:
            raise RuntimeError(f"GroupwiseIntQuantTensor not valid.")

    @staticmethod
    def check_input_type(tensor):
        if not isinstance(tensor, GroupwiseIntQuantTensor):
            raise RuntimeError("Tensor is not a GroupwiseIntQuantTensor")

    @staticmethod
    def is_zero_zero_point(tensor):
        GroupwiseIntQuantTensor.check_input_type(tensor)
        return (tensor.zero_point == 0.).all()

    def check_scaling_factors_same(self, other):
        if self.training:
            return True
        if not torch.allclose(self.scale, other.scale):
            raise RuntimeError("Scaling factors are different")

    def check_zero_points_same(self, other):
        if self.training:
            return True
        if not torch.allclose(self.zero_point, other.zero_point):
            raise RuntimeError("Zero points are different")

    def check_bit_width_same(self, other):
        if not torch.allclose(self.exponent_bit_width,
                              other.exponent_bit_width) and not torch.allclose(
                                  self.mantissa_bit_width, other.mantissa_bit_width):
            raise RuntimeError("Bit widths are different")

    def check_exponent_bias(self, other):
        if not torch.allclose(self.exponent_bias, other.exponent_bias):
            raise RuntimeError("Bit widths are different")

    def check_inf_nan_same(self, other):
        if not (set(self.inf_values) == set(other.inf_values)) and not (set(self.nan_values) == set(
                other.nan_values)):
            raise RuntimeError("Floating point representations are different")

    def check_sign_same(self, other):
        if not self.signed == other.signed:
            raise RuntimeError("Signs are different")

    def view(self, *args, **kwargs):
        return self.value.view(*args, **kwargs)  #self.set(value=self.value.view(*args, **kwargs))

    def reshape(self, *args, **kwargs):
        return self.value.reshape(
            *args, **kwargs)  # self.set(value=self.value.reshape(*args, **kwargs))

    def flatten(self, *args, **kwargs):
        return self.value.flatten(
            *args, **kwargs)  #self.set(value=self.value.flatten(*args, **kwargs))

    def transpose(self, *args, **kwargs):
        value = self.value.transpose(*args, **kwargs)
        return value

    def permute(self, *args, **kwargs):
        value = self.value.permute(*args, **kwargs)
        return value

    # Reference: https://docs.python.org/3/reference/datamodel.html#emulating-numeric-types

    def __neg__(self):
        neg_deq = -self.minifloat(float_datatype=True)
        _, scale, zp = self.expand()

        neg_value = (-neg_deq - zp) * scale
        # In case the dtype of self.minifloat is different from the one of the scale
        neg_value = neg_value.type(scale.dtype)
        neg_value = GroupwiseIntQuantTensor.from_expanded(
            neg_value, self.group_size, self.group_dim, compress=False)
        scale = GroupwiseIntQuantTensor.from_expanded(
            scale, self.group_size, self.group_dim, compress=True)
        if self.signed:
            return GroupwiseIntQuantTensor(
                value=neg_value,
                scale=scale,
                zero_point=self.zero_point,
                group_size=self.group_size,
                group_dim=self.group_dim,
                bit_width=self.bit_width,
                signed=self.signed,
                training=self.training,
                saturating=self.saturating)
        else:
            # TODO: implement
            raise NotImplementedError

    def __add__(self, other):
        if isinstance(other, QuantTensor):
            return self.value + other.value
        else:
            output = self.value + other
        return output

    def __mul__(self, other):
        if isinstance(other, QuantTensor):
            return self.value * other.value
        else:
            output = self.value * other
        return output

    def __str__(self):
        return f"GroupwiseIntQuantTensor(value={self.value}, scale={self.scale}, zero_point={self.zero_point}, group_size={self.group_size}, group_dim={self.group_dim}, bit_width={self.bit_width}, signed_t={self.signed_t}, training_t={self.training_t})"

    def __truediv__(self, other):
        if isinstance(other, QuantTensor):
            return self.value / other.value
        else:
            output = self.value / other
        return output

    def __abs__(self):
        if self.signed:
            neg_deq = self.minifloat(float_datatype=True)
            _, scale, zp = self.expand()

            # In case the dtype of self.minifloat is different from the one of the scale
            abs_value = (neg_deq - zp) * scale
            # In case the dtype of self.minifloat is different from the one of the scale
            abs_value = abs_value.type(scale.dtype)
            abs_value = GroupwiseIntQuantTensor.from_expanded(
                abs_value, self.group_size, self.group_dim, compress=False)
            scale = GroupwiseIntQuantTensor.from_expanded(
                scale, self.group_size, self.group_dim, compress=True)
            return GroupwiseIntQuantTensor(
                value=abs_value,
                scale=self.scale,
                zero_point=self.zero_point,
                group_size=self.group_size,
                group_dim=self.group_dim,
                bit_width=self.bit_width,
                signed=False,
                training=self.training,
                saturating=self.saturating)
        else:
            return self



# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

from abc import ABC
from abc import abstractmethod
from typing import Tuple

import torch

from brevitas.function.ops import max_float
from brevitas.function.ops import max_int
from brevitas.function.ops import min_int
from brevitas.proxy.float_parameter_quant import WeightFloatQuantProxyFromInjector
from brevitas.proxy.float_runtime_quant import ActFloatQuantProxyFromInjector
from brevitas.proxy.float_runtime_quant import ActFloatQuantProxyFromInjectorBase
from brevitas.proxy.parameter_quant import BiasQuantProxyFromInjector
from brevitas.proxy.parameter_quant import WeightQuantProxyFromInjector
from brevitas.proxy.runtime_quant import ActQuantProxyFromInjector
from brevitas.utils.torch_utils import float_internal_scale


class InferenceHandler(torch.nn.Module, ABC):

    def attach_debug_info(self, module):
        pass

    @abstractmethod
    def prepare_for_export(self, module):
        pass

    @abstractmethod
    def quantize(self, x):
        pass

    @abstractmethod
    def dequantize(self, x):
        pass


class IntInferencetHandler(InferenceHandler):
    handled_layer = (ActQuantProxyFromInjector, BiasQuantProxyFromInjector)

    def attach_debug_info(self, module):
        pass

    def prepare_for_export(self, module):
        if module.is_quant_enabled:
            self.scale = module.scale()
            self.zero_point = module.zero_point().to(self.scale.device)
            self.bit_width = module.bit_width()
            self.min_clamp = min_int(module.is_signed, module.is_narrow_range, self.bit_width)
            self.max_clamp = max_int(module.is_signed, module.is_narrow_range, self.bit_width)

    def quantize(self, x):
        return torch.clamp(
            torch.round(x / self.scale + self.zero_point), self.min_clamp, self.max_clamp)

    def dequantize(self, x):
        return (x - self.zero_point) * self.scale

    def forward(self, x, unused_scale=None) -> Tuple[torch.Tensor]:
        return self.dequantize(self.quantize(x)), self.scale, self.zero_point, self.bit_width


class IntWeightInferencetHandler(IntInferencetHandler):
    handled_layer = WeightQuantProxyFromInjector

    def prepare_for_export(self, module):
        if module.is_quant_enabled:
            self.cached_weight = None
            super().prepare_for_export(module)
            if module._cached_weight is not None and not module.cache_inference_quant_weight_metadata_only:
                self.cached_weight = module._cached_weight.value

    def forward(self, x) -> Tuple[torch.Tensor]:
        if self.cached_weight is not None:
            x = self.cached_weight
        else:
            x = self.dequantize(self.quantize(x))
        return x, self.scale, self.zero_point, self.bit_width


class FloatInferencetHandler(InferenceHandler):
    handled_layer = (ActFloatQuantProxyFromInjector, BiasQuantProxyFromInjector)

    def prepare_for_export(self, module):
        if module.is_quant_enabled:
            self.scale = module.scale()
            self.zero_point = module.zero_point().to(self.scale.device)
            self.exponent_bit_width = module.exponent_bit_width()
            self.mantissa_bit_width = module.mantissa_bit_width()
            self.exponent_bias = module.exponent_bias()
            self.saturating = module.is_saturating()
            self.inf_values = module.inf_values()
            self.nan_values = module.nan_values()
            self.eps = torch.finfo(self.scale.dtype).tiny
            if hasattr(module.tensor_quant, 'float_to_int_impl'):
                self.float_to_int_impl = module.tensor_quant.float_to_int_impl
                self.float_clamp_impl = module.tensor_quant.float_clamp_impl
            elif hasattr(module, 'fused_activation_quant_proxy'):
                self.float_to_int_impl = module.fused_activation_quant_proxy.tensor_quant.float_to_int_impl
                self.float_clamp_impl = module.fused_activation_quant_proxy.tensor_quant.float_clamp_impl

            self.max_clamp = max_float(
                self.exponent_bit_width, self.mantissa_bit_width, self.exponent_bias)
            self.min_clamp = -self.max_clamp
            self.fp_internal_scale_min = 1. - self.exponent_bias - self.mantissa_bit_width
            self.max_value = max_float(
                self.exponent_bit_width, self.mantissa_bit_width, self.exponent_bias)
            self.min_value = torch.tensor(0.) if not module.is_signed else -self.max_value

    def quantize(self, x):
        # Compute masks
        inf_mask = x.isinf()
        p_max_val_mask = x > self.max_value
        n_max_val_mask = -x > self.max_value

        # Quantize
        x = x / self.scale
        internal_scale = float_internal_scale(
            x, self.mantissa_bit_width, self.fp_internal_scale_min, self.eps)
        x = internal_scale * self.float_to_int_impl(x / internal_scale)

        # Clamp
        x = self.float_clamp_impl.saturating_clamp(x, self.max_value, self.min_value)
        if not self.saturating:
            x = self.float_clamp_impl.inf_nan_clamp(x, inf_mask, p_max_val_mask, n_max_val_mask)

        return x

    def dequantize(self, x):
        return (x - self.zero_point) * self.scale

    def forward(self, x) -> Tuple[torch.Tensor]:
        return self.dequantize(self.quantize(x)), self.scale, self.zero_point, self.exponent_bit_width, self.mantissa_bit_width, self.exponent_bias, self.saturating, self.inf_values, self.nan_values


class FloatWeightInferencetHandler(FloatInferencetHandler):
    handled_layer = WeightFloatQuantProxyFromInjector

    def prepare_for_export(self, module):
        if module.is_quant_enabled:
            self.cached_weight = None
            super().prepare_for_export(module)
            if module._cached_weight is not None and not module.cache_inference_quant_weight_metadata_only:
                self.cached_weight = module._cached_weight.value

    def forward(self, x) -> Tuple[torch.Tensor]:
        if self.cached_weight is not None:
            x = self.cached_weight
        else:
            x = self.dequantize(self.quantize(x))
        return x, self.scale, self.zero_point, self.exponent_bit_width, self.mantissa_bit_width, self.exponent_bias, self.saturating, self.inf_values, self.nan_values



"""
Copyright (C) 2024, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
"""

import torch
from tqdm import tqdm

from brevitas.graph.gpfq import gpfq_mode
from brevitas.graph.gptq import gptq_mode


@torch.no_grad()
def apply_gptq(
        model,
        dataloader,
        act_order=True,
        group_of_parallel_layers=None,
        use_quant_activations=True,
        create_weight_orig=False):
    with gptq_mode(model,
                   act_order=act_order,
                   group_of_parallel_layers=group_of_parallel_layers,
                   use_quant_activations=use_quant_activations,
                   create_weight_orig=create_weight_orig) as gptq:
        gptq_model = gptq.model
        for _ in tqdm(range(gptq.num_layers)):
            for inps in dataloader:
                gptq_model(**inps)
            gptq.update()


@torch.no_grad()
def apply_gpfq(model, dataloader, act_order=True, group_of_parallel_layers=None):
    with gpfq_mode(model, act_order=act_order,
                   group_of_parallel_layers=group_of_parallel_layers) as gpfq:
        gpfq_model = gpfq.model
        for _ in tqdm(range(gpfq.num_layers)):
            for inps in dataloader:
                gpfq_model(**inps)
            gpfq.update()



"""
Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
SPDX-License-Identifier: MIT
"""

import json
import os

from safetensors.torch import save_file
import torch
from torch import nn

from brevitas.nn.equalized_layer import EqualizedModule
from brevitas.nn.quant_layer import QuantNonLinearActLayer
from brevitas.nn.quant_layer import QuantWeightBiasInputOutputLayer
from brevitas_examples.llm.llm_quant.export import brevitas_proxy_export_mode


def export_onnx(pipe, trace_inputs, output_dir, export_manager):
    output_path = os.path.join(output_dir, 'unet.onnx')
    print(f"Saving unet to {output_path} ...")
    with torch.no_grad(), brevitas_proxy_export_mode(pipe.unet, export_manager):
        torch.onnx.export(pipe.unet, args=trace_inputs, f=output_path)


def handle_quant_param(layer, layer_dict):
    input_scale = layer.input_quant.export_handler.symbolic_kwargs['dequantize_symbolic_kwargs'][
        'scale'].data
    input_zp = layer.input_quant.export_handler.symbolic_kwargs['dequantize_symbolic_kwargs'][
        'zero_point'].data
    weight_scale = layer.weight_quant.export_handler.symbolic_kwargs['dequantize_symbolic_kwargs'][
        'scale'].data
    weight_zp = layer.weight_quant.export_handler.symbolic_kwargs['dequantize_symbolic_kwargs'][
        'zero_point'].data
    if layer.output_quant.export_handler.symbolic_kwargs is not None:
        output_scale = layer.output_quant.export_handler.symbolic_kwargs[
            'dequantize_symbolic_kwargs']['scale'].data

        layer_dict['output_scale'] = output_scale.cpu().numpy().tolist()
        layer_dict['output_scale_shape'] = output_scale.shape
    layer_dict['input_scale'] = input_scale.cpu().numpy().tolist()
    layer_dict['input_scale_shape'] = input_scale.shape
    layer_dict['input_zp'] = input_zp.to(torch.float32).cpu().numpy().tolist()
    layer_dict['input_zp_shape'] = input_zp.shape
    layer_dict['input_zp_dtype'] = str(input_zp.dtype)
    layer_dict['weight_scale'] = weight_scale.cpu().numpy().tolist()
    nelems = layer.weight.shape[0]
    weight_scale_shape = [nelems] + [1] * (layer.weight.data.ndim - 1)
    layer_dict['weight_scale_shape'] = weight_scale_shape
    if torch.sum(weight_zp.to(torch.float32)) != 0.:
        weight_zp = weight_zp - 128.  # apply offset to have signed z
        layer_dict['weight_zp'] = weight_zp.to(torch.float32).cpu().numpy().tolist()
        layer_dict['weight_zp_shape'] = weight_scale_shape
        layer_dict['weight_zp_dtype'] = str(weight_zp.dtype)
    return layer_dict


def export_quant_params(pipe, output_dir, export_vae=False):
    quant_output_path = os.path.join(output_dir, 'quant_params.json')
    unet_output_path = os.path.join(output_dir, 'params.safetensors')
    print(f"Saving unet to {unet_output_path} ...")
    if export_vae:
        vae_output_path = os.path.join(output_dir, 'vae.safetensors')
        print(f"Saving vae to {vae_output_path} ...")
    from brevitas.export.onnx.standard.qcdq.manager import StdQCDQONNXManager
    export_manager = StdQCDQONNXManager
    export_manager.change_weight_export(
        export_weight_q_node=True)  # We're exporting FP weights + quantization parameters
    quant_params = dict()
    state_dict = pipe.unet.state_dict()
    state_dict = {k: v for (k, v) in state_dict.items() if 'tensor_quant' not in k}
    state_dict = {k: v for (k, v) in state_dict.items() if not k.endswith('.scale.weight')}
    state_dict = {k.replace('.layer.', '.'): v for (k, v) in state_dict.items()}

    handled_quant_layers = set()
    with torch.no_grad(), brevitas_proxy_export_mode(pipe.unet, export_manager):
        for name, module in pipe.unet.named_modules():
            if isinstance(module, EqualizedModule):
                if id(module.layer) in handled_quant_layers:
                    raise RuntimeError("This should not happen")
                if isinstance(module.layer, QuantWeightBiasInputOutputLayer):
                    layer_dict = dict()
                    full_name = name
                    smoothquant_param = module.scale.weight

                    layer_dict['smoothquant_mul'] = smoothquant_param.data.cpu().numpy().tolist()
                    layer_dict['smoothquant_mul_shape'] = module.scale.runtime_shape
                    layer_dict = handle_quant_param(module.layer, layer_dict)

                    quant_params[full_name] = layer_dict
                    handled_quant_layers.add(id(module.layer))
                else:
                    layer_dict = dict()
                    full_name = name
                    smoothquant_param = module.scale.weight

                    layer_dict['smoothquant_mul'] = smoothquant_param.data.cpu().numpy().tolist()
                    layer_dict['smoothquant_mul_shape'] = module.scale.runtime_shape
                    quant_params[full_name] = layer_dict
                    handled_quant_layers.add(id(module.layer))
            elif isinstance(
                    module,
                    QuantWeightBiasInputOutputLayer) and id(module) not in handled_quant_layers:
                full_name = name
                layer_dict = dict()
                layer_dict = handle_quant_param(module, layer_dict)
                quant_params[full_name] = layer_dict
                handled_quant_layers.add(id(module))
            elif isinstance(module, QuantNonLinearActLayer):
                full_name = name
                layer_dict = dict()
                act_scale = module.act_quant.export_handler.symbolic_kwargs[
                    'dequantize_symbolic_kwargs']['scale'].data
                act_zp = module.act_quant.export_handler.symbolic_kwargs[
                    'dequantize_symbolic_kwargs']['zero_point'].data
                layer_dict['act_scale'] = act_scale.cpu().numpy().tolist()
                layer_dict['act_scale_shape'] = act_scale.shape
                layer_dict['act_zp'] = act_zp.to(torch.float32).cpu().numpy().tolist()
                layer_dict['act_zp_shape'] = act_zp.shape
                layer_dict['act_zp_dtype'] = str(act_zp.dtype)
                quant_params[full_name] = layer_dict
                handled_quant_layers.add(id(module))
    with open(quant_output_path, 'w') as file:
        json.dump(quant_params, file, indent="  ")
    save_file(state_dict, unet_output_path)
    if export_vae:
        save_file(pipe.vae.state_dict(), vae_output_path)



# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

from abc import ABCMeta
from abc import abstractmethod
from inspect import isclass
import math
from typing import Optional, Tuple, Union
from warnings import warn

import packaging.version
import torch
from torch import nn
from torch import Tensor
import torch.jit
from torch.nn.utils.rnn import PackedSequence

from brevitas import config
from brevitas import is_dynamo_compiling
from brevitas import torch_version
from brevitas.common import ExportMixin
from brevitas.inject import ExtendedInjector
from brevitas.inject import Injector
from brevitas.quant_tensor import _unpack_quant_tensor
from brevitas.quant_tensor import FloatQuantTensor
from brevitas.quant_tensor import IntQuantTensor
from brevitas.quant_tensor import QuantTensor
from brevitas.quant_tensor.groupwise_float_quant_tensor import GroupwiseFloatQuantTensor
from brevitas.quant_tensor.groupwise_int_quant_tensor import GroupwiseIntQuantTensor

from .utils import filter_kwargs


class QuantProxyMixin(object):
    __metaclass__ = ABCMeta

    def __init__(
            self,
            quant,
            proxy_protocol,
            none_quant_injector,
            proxy_prefix: str,
            kwargs_prefix: str,
            **kwargs):
        proxy_name = proxy_prefix + 'quant'
        if quant is None:
            quant_injector = none_quant_injector.let(**filter_kwargs(kwargs_prefix, kwargs))
            quant = quant_injector.proxy_class(self, quant_injector)
        elif isclass(quant) and issubclass(quant, (Injector, ExtendedInjector)):
            quant_injector = quant
            quant_injector = quant_injector.let(**filter_kwargs(kwargs_prefix, kwargs))
            quant = quant_injector.proxy_class(self, quant_injector)
        else:
            if not isinstance(quant, proxy_protocol):
                raise RuntimeError(
                    "The quantizer passed does not adhere to the quantization protocol.")
            quant.add_tracked_module(self)
            if filter_kwargs(kwargs_prefix, kwargs):
                warn('Keyword arguments are being passed but they not being used.')
        setattr(self, proxy_name, quant)


class QuantLayerMixin(ExportMixin):
    __metaclass__ = ABCMeta

    def __init__(self, return_quant_tensor: bool):
        ExportMixin.__init__(self)
        self.accept_quant_tensor = True
        self.return_quant_tensor = return_quant_tensor

    @property
    @abstractmethod
    def channelwise_separable(self) -> bool:
        pass

    def get_quant_tensor_class(self, inp: Union[Tensor, QuantTensor]):
        quant_tensor_classes = [
            IntQuantTensor, FloatQuantTensor, GroupwiseIntQuantTensor, GroupwiseFloatQuantTensor]
        for qt_class in quant_tensor_classes:
            if len(inp) == len(qt_class._fields):
                return qt_class
        return None

    def unpack_input(self, inp: Union[Tensor, QuantTensor]) -> Union[Tensor, QuantTensor]:
        # Hack to recognize a QuantTensor that has decayed to a tuple
        # when used as input to tracing (e.g. during ONNX export)
        if (torch._C._get_tracing_state() is not None and isinstance(inp, tuple) and
                all([isinstance(t, Tensor) for t in inp])):
            qt_class = self.get_quant_tensor_class(inp)
            if qt_class is not None:
                inp = qt_class(*inp)
        if not torch._C._get_tracing_state() and not is_dynamo_compiling():
            if isinstance(inp, QuantTensor):
                inp = inp.set(value=inp.value.rename(None))
            else:
                inp = inp.rename(None)
        return inp

    def pack_output(self, quant_output: Union[Tensor, QuantTensor]) -> Union[Tensor, QuantTensor]:
        if self.return_quant_tensor:
            assert isinstance(quant_output, QuantTensor), 'QuantLayer is not correctly configured, check if warnings were raised'
            return quant_output
        else:
            return _unpack_quant_tensor(quant_output)


class QuantRecurrentLayerMixin(ExportMixin):
    __metaclass__ = ABCMeta

    def __init__(
            self,
            cell: nn.Module,
            io_quant: nn.Module,
            input_size: int,
            hidden_size: int,
            reverse_input: bool,
            quantize_output_only: bool,
            shared_input_hidden_weights: bool,
            return_quant_tensor: bool):
        ExportMixin.__init__(self)
        self.cell = cell
        self.io_quant = io_quant
        self.input_size = input_size
        self.hidden_size = hidden_size
        self.reverse_input = reverse_input
        self.quantize_output_only = quantize_output_only
        self.shared_input_hidden_weights = shared_input_hidden_weights
        self.return_quant_tensor = return_quant_tensor
        self.accept_quant_tensor = True
        self.fast_mode = True
        self._fast_cell = None

    @property
    @abstractmethod
    def weights_to_share(self):
        pass

    @property
    @abstractmethod
    def quantizers_to_share(self):
        pass

    @property
    @abstractmethod
    def fast_cell(self):
        pass

    @property
    def requires_export_handler(self):
        return True

    def state_dict(self, destination=None, prefix='', keep_vars=False):
        output_dict = super(QuantRecurrentLayerMixin, self).state_dict(
            destination=destination, prefix=prefix, keep_vars=keep_vars)
        for key in list(output_dict.keys()):
            if '_fast_cell' in key:
                del output_dict[key]
        return output_dict

    @staticmethod
    def gate_params_fwd(gate, quant_input):
        quant_weight_ih = gate.input_weight()
        quant_weight_hh = gate.hidden_weight()
        quant_bias = gate.bias_quant(gate.bias, quant_input, quant_weight_ih)
        return quant_weight_ih, quant_weight_hh, quant_bias

    def reset_parameters(self) -> None:
        stdv = 1.0 / math.sqrt(self.hidden_size)
        for name, weight in self.named_parameters():
            if 'gate' in name:
                nn.init.uniform_(weight, -stdv, stdv)

    def maybe_quantize_input(self, inp):
        if isinstance(inp, PackedSequence):
            raise RuntimeError("PackedSequence input currently not supported.")
        quant_input = inp
        if not self.quantize_output_only:
            quant_input = self.io_quant(quant_input)
        return quant_input

    def maybe_quantize_state(self, inp, state, quant):
        if state is None:
            batch_size = inp.size(0) if self.cell.batch_first else inp.size(1)
            quant_state = torch.zeros(
                int(batch_size), self.hidden_size, dtype=inp.dtype, device=inp.device)
        else:
            quant_state = quant(state)
        return quant_state

    def pack_quant_outputs(self, quant_outputs):
        # In export mode, quant_outputs has the shape of the output concatenated value
        # Even though we check that return_quant_tensor can be enabled only with io_quant != None,
        # inner layers in a deep network overrides it, so we check again.
        if self.export_mode:
            if self.return_quant_tensor and self.io_quant.is_quant_enabled:
                return IntQuantTensor(
                    quant_outputs,
                    self.io_quant.scale(),
                    self.io_quant.zero_point(),
                    self.io_quant.bit_width(),
                    self.io_quant.is_signed,
                    self.training)
            else:
                return quant_outputs
        seq_dim = 1 if self.cell.batch_first else 0
        if self.return_quant_tensor and self.io_quant.is_quant_enabled:
            outputs = [
                IntQuantTensor(
                    torch.unsqueeze(quant_output[0], dim=seq_dim),
                    quant_output[1],
                    quant_output[2],
                    quant_output[3],
                    self.io_quant.is_signed,
                    self.training) for quant_output in quant_outputs]
        else:
            outputs = [torch.unsqueeze(o[0], dim=seq_dim) for o in quant_outputs]
        if self.reverse_input:
            return torch.cat(list(reversed(outputs)), dim=seq_dim)
        else:
            return torch.cat(outputs, dim=seq_dim)

    def pack_quant_state(self, quant_state, quant):
        # Even though we check that return_quant_tensor can be enabled only with quant != None,
        # inner layers in a deep network overrides it, so we check again.
        if self.export_mode:
            if self.return_quant_tensor and quant.is_quant_enabled:
                quant_state = IntQuantTensor(
                    torch.unsqueeze(quant_state, dim=0),
                    quant.scale(),
                    quant.zero_point(),
                    quant.bit_width(),
                    quant.is_signed,
                    self.training)
            else:
                quant_state = torch.unsqueeze(quant_state, dim=0)
        else:
            if self.return_quant_tensor and quant.is_quant_enabled:
                quant_state = IntQuantTensor(
                    torch.unsqueeze(quant_state[0], dim=0),
                    quant_state[1],
                    quant_state[2],
                    quant_state[3],
                    quant.is_signed,
                    self.training)
            else:
                quant_state = torch.unsqueeze(quant_state[0], dim=0)
        return quant_state

    def _wrap_act_proxy(self, quant_name):

        class _Wrapper(nn.Module):

            def __init__(self, module_to_wrap=None):
                super(_Wrapper, self).__init__()
                if module_to_wrap is None:
                    module_to_wrap = nn.Identity()
                self.module_to_wrap = module_to_wrap

            def forward(
                self, x: torch.Tensor
            ) -> Tuple[Tensor, Optional[Tensor], Optional[Tensor], Optional[Tensor]]:
                x = self.module_to_wrap(x)
                return (x, None, None, None)

        proxy = getattr(self.cell, quant_name)
        if proxy.fused_activation_quant_proxy is None:
            proxy = _Wrapper(proxy.fused_activation_quant_proxy)
        else:
            proxy = proxy.fused_activation_quant_proxy
        return proxy



# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

from .dataset import *
from .evaluate import *
from .export import *
from .train import *



# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

from functools import wraps

from .onnx.debug import enable_debug
from .onnx.qonnx.manager import QONNXManager
from .onnx.standard.qcdq.manager import StdQCDQONNXManager
from .torch.qcdq.manager import TorchQCDQManager


@wraps(QONNXManager.export)
def export_brevitas_onnx(*args, **kwargs):  # alias for qonnx
    return QONNXManager.export(*args, **kwargs)


@wraps(QONNXManager.export)
def export_qonnx(*args, **kwargs):
    return QONNXManager.export(*args, **kwargs)


@wraps(StdQCDQONNXManager.export)
def export_onnx_qcdq(*args, **kwargs):
    return StdQCDQONNXManager.export(*args, **kwargs)


@wraps(TorchQCDQManager.export)
def export_torch_qcdq(*args, **kwargs):
    return TorchQCDQManager.export(*args, **kwargs)



"""
Adapted from https://github.com/IST-DASLab/gptq, released under the following LICENSE:

Copyright 2023 IST-DASLab

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import torch
from torch import nn
from tqdm import tqdm


def create_validation_dataloader(data, seqlen, device):
    nsamples = data['input_ids'].numel() // seqlen
    val_dataloader = []
    for i in tqdm(range(nsamples)):
        batch = data['input_ids'][:, (i * seqlen):((i + 1) * seqlen)].to(device)
        attention_mask = torch.ones_like(batch)
        val_dataloader.append({'input_ids': batch, 'attention_mask': attention_mask})
    return val_dataloader


@torch.no_grad()
def model_eval(model, valenc, seqlen):
    nsamples = len(valenc)
    with torch.no_grad():
        nlls = []
        for inps in valenc:
            lm_logits = model(**inps)['logits']
            shift_logits = lm_logits[:, :-1, :].contiguous()
            dev = shift_logits.device
            shift_labels = inps['input_ids'][:, 1:].to(dev)
            shift_logits = shift_logits.to(dev)
            loss_fct = nn.CrossEntropyLoss()
            loss = loss_fct(shift_logits.view(-1, shift_logits.size(-1)), shift_labels.view(-1))
            neg_log_likelihood = loss.float() * seqlen
            nlls.append(neg_log_likelihood)
        ppl = torch.exp(torch.stack(nlls).sum() / (nsamples * seqlen))
    return ppl



"""
Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
"""

# yapf: disable
from typing import List, Tuple

import torch
import torch.utils.cpp_extension
import torch_mlir
from torch_mlir.jit_ir_importer.build_tools.registry import _rename_python_keyword_parameter_name
from torch_mlir.jit_ir_importer.build_tools.registry import JitOperator
from torch_mlir.jit_ir_importer.build_tools.registry import SIG_ATTR_TYPE

from brevitas.backport.fx._symbolic_trace import wrap


def patched_has_value_semantics_function_signature(self):
    """Gets the Python function signature for this op's has_value_semantics function.
    While this is technically debug-only output, it is useful to copy-paste
    it from the debug dump into the library definitions, as many
    ops have extra default arguments and stuff that are tedious to write out
    right.
    """

    def parameter_decl_builder(arg: "SIG_ATTR_TYPE") -> str:
        parameter_name = _rename_python_keyword_parameter_name(arg["name"])
        return f"{parameter_name}"

    def ret_decl_builder(arg: "SIG_ATTR_TYPE") -> str:
        return "None"

    return self._get_function_signature(
        "has_value_semantics", parameter_decl_builder, ret_decl_builder)


JitOperator.get_has_value_semantics_function_signature = patched_has_value_semantics_function_signature


def matmul_rhs_group_quant(
        lhs: torch.Tensor,
        rhs: torch.Tensor,
        rhs_scale: torch.Tensor,
        rhs_zero_point: torch.Tensor,
        rhs_bit_width: int,
        rhs_group_size: int):
    # This is just a placeholder for the actual implementation that provides correct shape/device/dtype
    if len(lhs.shape) == 3 and len(rhs.shape) == 2:
        return torch.randn(
            lhs.shape[0], lhs.shape[1], rhs.shape[0], device=lhs.device, dtype=lhs.dtype)
    elif len(lhs.shape) == 2 and len(rhs.shape) == 2:
        return torch.randn(lhs.shape[0], rhs.shape[0], device=lhs.device, dtype=lhs.dtype)
    else:
        raise ValueError("Input shapes not supported.")


brevitas_lib = torch.library.Library("quant", "DEF")
brevitas_lib.define(
    "matmul_rhs_group_quant(Tensor lhs, Tensor rhs, Tensor rhs_scale, Tensor rhs_zero_point, int rhs_bit_width, int rhs_group_size) -> Tensor"
)
brevitas_lib.impl("matmul_rhs_group_quant", matmul_rhs_group_quant)


def quant〇matmul_rhs_group_quant〡shape(lhs: List[int], rhs: List[int], rhs_scale: List[int], rhs_zero_point: List[int], rhs_bit_width: int, rhs_group_size: int) -> List[int]:
    if len(lhs) == 3 and len(rhs) == 2:
        return [lhs[0], lhs[1], rhs[0]]
    elif len(lhs) == 2 and len(rhs) == 2:
        return [lhs[0], rhs[0]]
    else:
        raise ValueError("Input shapes not supported.")


def quant〇matmul_rhs_group_quant〡dtype(lhs_rank_dtype: Tuple[int, int], rhs_rank_dtype: Tuple[int, int], rhs_scale_rank_dtype: Tuple[int, int], rhs_zero_point_rank_dtype: Tuple[int, int], rhs_bit_width: int, rhs_group_size: int) -> int:
    # output dtype is the dtype of the lhs float input
    lhs_rank, lhs_dtype = lhs_rank_dtype
    return lhs_dtype


def quant〇matmul_rhs_group_quant〡has_value_semantics(lhs, rhs, rhs_scale, rhs_zero_point, rhs_bit_width, rhs_group_size) -> None:
    return
# yapf: enable

brevitas_matmul_rhs_group_quant_library = [
    quant〇matmul_rhs_group_quant〡shape,
    quant〇matmul_rhs_group_quant〡dtype,
    quant〇matmul_rhs_group_quant〡has_value_semantics]

if __name__ == '__main__':

    class CustomOpExampleModule(torch.nn.Module):

        def __init__(self):
            super().__init__()

        def forward(
                self: torch.nn.Module,
                lhs: torch.Tensor,
                rhs: torch.Tensor,
                rhs_scale: torch.Tensor,
                rhs_zero_point: torch.Tensor):
            return torch.ops.quant.matmul_rhs_group_quant(
                lhs, rhs, rhs_scale, rhs_zero_point, rhs_bit_width=8, rhs_group_size=128)

    mod = CustomOpExampleModule()
    mod.eval()

    module = torch_mlir.compile(
        mod, (torch.ones(3, 4), torch.ones(5, 4), torch.ones(1), torch.ones(1)),
        output_type="torch",
        backend_legal_ops=["quant.matmul_rhs_group_quant"],
        extra_library=brevitas_matmul_rhs_group_quant_library)
    print(module)
